"""
This code performs name clustering on PROP PER mentions, grouping together different proper names only that refer to the same individual
e.g., Tom, Tom Saywer, Mr. Tom Sawyer, Mr. Sawyer -> TOM SAYWER

"""

from collections import Counter
import sys
import itertools
import pkg_resources

class NameCoref:

	def __init__(self, aliasFile):
		self.honorifics={"mr":1, "mr.":1, "mrs":1, "mrs.":1, "miss":1, "uncle":1, "aunt":1, "lady":1, "lord":1, "monsieur":1, "master":1, "mistress":1}
		self.aliases={}
		with open(aliasFile) as file:
			for line in file:
				cols=line.rstrip().split("\t")
				canonical=cols[0]
				nicknames=cols[1:]
				for nickname in nicknames:

					if nickname.lower() not in self.aliases:
						self.aliases[nickname.lower()]={}
					self.aliases[nickname.lower()][canonical.lower()]=1

	def get_variants(self, parts):
		variants={}
		for i in range(len(parts)):
			if parts[i].lower() not in self.honorifics:
				variants[parts[i]]=1

			for j in range(i+1, len(parts)):
				variants["%s %s" % (parts[i], parts[j])]=1

				for k in range(j+1, len(parts)):
					variants["%s %s %s" % (parts[i], parts[j], parts[k])]=1

					for l in range(k+1, len(parts)):
						variants["%s %s %s %s" % (parts[i], parts[j], parts[k], parts[l])]=1

						for m in range(l+1, len(parts)):
							variants["%s %s %s %s %s" % (parts[i], parts[j], parts[k], parts[l], parts[m])]=1

							for n in range(m+1, len(parts)):
								variants["%s %s %s %s %s %s" % (parts[i], parts[j], parts[k], parts[l], parts[m], parts[n])]=1

								for o in range(n+1, len(parts)):
									variants["%s %s %s %s %s %s %s" % (parts[i], parts[j], parts[k], parts[l], parts[m], parts[n], parts[o])]=1
								


		return variants

	def get_canonical(self, name_tokens):

		"""
		Given a alias dictionary that maps aliases (nicknames, alternative names) to a "canonical" version 
		of the name, return all possible canonical versions of the input

		Alias dictionary:

		Em -> Emily
		Em -> Emma
		The Great Bambino -> Babe Ruth


		Input -> Output

		Em -> [["Emily"], ["Emma"]]
		Em Smith -> [["Emily", "Smith"], ["Emma", "Smith"]]
		The Great Bambino -> [["Babe", "Ruth"]]

		"""

		# first, if a name is a complete match for an alias, just return the canonicals for the alias
		# Em -> ["Emily", "Emma"]
		# The Great Bambino -> ["Babe Ruth"]
		name=' '.join(name_tokens).lower()
		if name in self.aliases:
			vals=[]
			for can in self.aliases[name]:
				vals.append(can.split(" "))
			return vals

		# next, check if any individual part of a name is an alias
		# Em Smith -> ["Emily Smith", "Emma Smith"]
		parts=[]
		for tok in name_tokens:
			if tok.lower() in self.aliases:
				parts.append(list(self.aliases[tok.lower()]))
			else:
				parts.append([tok])
		canonicals=[]

		canonicals=[]
		for i in itertools.product(*parts): 
			canonicals.append(list(i))

		return canonicals

	def name_cluster(self, entities, is_named, existing_refs):

		"""
		Get counts of all unique names (to be used for disambiguation later)
		"""

		uniq=Counter()
		for i, val in enumerate(is_named):
			if val == 1:
				
				# only consider names with fewer than 10 tokens
				# (longer are likely errors and significantly slow down processing)
				if len(entities[i]) < 10:
					name=' '.join(entities[i]).lower()
					if name != "":
						uniq[name]+=1

		
		"""

		Remove names that are complete subsets of others
		
		e.g., if uniq = ["Em Smith", "Em", "Emma Smith", "Tom", "Tom Sawyer"], then remove "Em", "Em Smith" and "Tom".
			* "Tom" is a subset of "Tom Sawyer"
			* "Em" -> "Emma" and "Emily", and "Emma" is a subset of "Emma Smith"
			* "Em Smith" -> "Emma Smith"
		"""

		subsets={}
		for name1 in uniq:
			canonicals1=self.get_canonical(name1.split(" "))
			for canonical1 in canonicals1:
				name1set=set(canonical1)
				for name2 in uniq:

					if name1 == name2:
						continue

					canonicals=self.get_canonical(name2.split(" "))
					for canonical in canonicals:

						name2set=set(canonical)

						if ' '.join(name1set) == ' '.join(name2set):
							continue

						if name1set.issuperset(name2set):
							subsets[name2]=1

		name_subpart_index={}

		"""

		Now map each possible ordered sub-permutation of a canonical name  (from length 1 to n) to its canonical version.
		
		e.g. the canonical name "David Foster Wallace" -> 
		"David Foster Wallace", "David Wallace", "Foster Wallace", "David Foster", "Foster Wallace", "David", "Foster", "Wallace"

		So if we see any of those phrases as names, we know they could refer to the entity "David Foster Wallace"

		This excludes honorifics like "Mr." and "Mrs." from being their own (unigram) variant, but they can appear in other permutations

		e.g. "Mr. Tom Sawyer" ->
		"Mr. Tom Sawyer", "Mr. Tom", "Mr. Sawyer", "Tom Sawyer", "Tom", "Sawyer"
		
		"""

		for ni, name in enumerate(uniq):
			if name in subsets:
				continue

			canonicals=self.get_canonical(name.split(" "))
			for canonical in canonicals:
				variants=self.get_variants(canonical)

				for v in variants:
					if v not in name_subpart_index:
						name_subpart_index[v]={}

					name_subpart_index[v][name]=1


		"""

		Now let's assign each name *mention* to its entity.

		Starting from the first entity to last, we'll look up the possible entities that the canonical version of
		this mention can be a variant of and assign it to the entity that has the highest score.

		The score for an entity is initialized as the mention count of the maximal name 
		("David Foster Wallace" or "Mr. Tom Sawyer" above), but is primarily driven by recency (with the score
		mainly being the location of the last assigned mention of that entity)

		"""

		charids={}
		max_id=1

		if len(existing_refs) > 0:
			max_id=max(existing_refs)+1

		lastSeen={}
		refs=[]
		for i, val in enumerate(is_named):

			if existing_refs[i] != -1:
				refs.append(existing_refs[i])
				continue

			if val == 1:

				canonicals=self.get_canonical(entities[i])
				name=' '.join(entities[i]).lower()

				top=None
				max_score=0

				for canonical in canonicals:
					canonical_name=' '.join(canonical).lower()

					if canonical_name in name_subpart_index:
						for entity in name_subpart_index[canonical_name]:
							score=uniq[entity]
							if entity in lastSeen:
								score+=lastSeen[entity]
							if score > max_score:
								max_score=score
								top=entity

				if top is not None:
					lastSeen[top]=i

					if top not in charids:
						charids[top]=max_id
						max_id+=1
					refs.append(charids[top])

				# this happens if the name is too long (longer than 7 words)
				else:
					refs.append(-1)
			else:
				refs.append(-1)
		return refs

	def calc_overlap(self, small, big):

		overlap=0
		for name in small:
			if name in big:
				overlap+=small[name]

		return overlap/sum(small.values())

	def read_file(self, spanFile):

		entities=[]
		is_named=[]
		with open(spanFile) as file:
			for line in file:
				cols=line.rstrip().split("\t")

				cat=cols[2].split("_")[1]
				prop=cols[2].split("_")[0]
				if prop != "PROP":
					continue
				text=cols[0].split(" ")
				lemma=cols[1]
				entity_pos=cols[3].split(" ")
				if cat == "PER":
					name_filt_pos=[]
					for pidx, pos in enumerate(entity_pos):
						if pos == "NOUN" or pos == "PROPN":
							name_filt_pos.append(text[pidx])

					entities.append(name_filt_pos)
					is_named.append(1)

		return entities, is_named

	def process(self, spanFile):

		entities, is_named=read_file(spanFile)
		cluster(entities, is_named)

	def cluster_identical_propers(self, entities, refs):

		""" Assign all mentions that are identical to the same entity (used in combination with only performing 
		pronominal coreference resolution) """

		max_id=1
		if len(refs) > 0:
			max_id=max(refs)+1

		names={}

		for idx, (s, e, full_cat, name) in enumerate(entities):
			prop=full_cat.split("_")[0]
			cat=full_cat.split("_")[1]
			
			if prop == "PROP" and cat != "PER":
				n=name.lower()
				key="%s_%s_%s" % (n, prop, cat)
				if key not in names:
					names[key]=max_id
					max_id+=1
				refs[idx]=names[key]

		return refs


	def cluster_noms(self, entities, refs):

		""" Assign all nominal mentions that are identical to the same entity (used in combination with only performing 
		pronominal coreference resolution) """

		names={}
		mapper={}

		for idx, (s, e, cat, name) in enumerate(entities):
			prop=cat.split("_")[0]
			if prop == "NOM":
				n=name.lower()
				if n not in names:
					names[n]=refs[idx]
				else:
					mapper[refs[idx]]=names[n]


		for idx, ref in enumerate(refs):
			if ref in mapper:
				refs[idx]=mapper[ref]

		return refs



	def cluster_narrator(self, entities, in_quotes, tokens):

		""" Create an entity for the first-person narrator from all mentiosn of "I", "me", "my" and "myself" outside of quotes """

		narrator_pronouns=set(["i", "me", "my", "myself"])
		refs=[]
		for idx, (s, e, _, name) in enumerate(entities):
			if in_quotes[idx] == 0 and name.lower() in narrator_pronouns:
				refs.append(0)
				# window=25
				# start=max(0, s-window)
				# end=min(e+25, len(tokens))
				# context=[tok.text for tok in tokens[start:end]]
				# print("narrator\t\t", name, ' '.join(context)) 
			else:
				refs.append(-1)
		return refs

	def cluster_only_nouns(self, entities, refs, tokens):

		hon_mapper={"mister":"mr.", "mr.":"mr.", "mr":"mr.", "mistah":"mr.", "mastah":"mr.", "master":"mr.",
		"miss":"miss", "ms.": "miss", "ms":"miss","missus":"miss","mistress":"miss",
		"mrs.":"mrs.", "mrs":"mrs."
		}

		def map_honorifics(term):
			term=term.lower()
			if term in hon_mapper:
				return hon_mapper[term]
			return None

		is_named=[]
		entity_names=[]

		for start, end, cat, text in entities:
			ner_prop=cat.split("_")[0]
			ner_type=cat.split("_")[1]
			if ner_prop == "PROP" and ner_type == "PER":
				is_named.append(1)
			else:
				is_named.append(0)

			new_text=[]
			for i in range(start,end+1):
				hon_mapped=map_honorifics(tokens[i].text)

				if (hon_mapped is not None or (tokens[i].pos == "NOUN" or tokens[i].pos == "PROPN")) and tokens[i].text.lower()[0] != tokens[i].text[0]:
					val=tokens[i].text
					if hon_mapped is not None:
						val=hon_mapped
					new_text.append(val)

			if len(new_text) > 0:
				entity_names.append(new_text)
			else:
				entity_names.append(text.split(" "))

		return self.cluster(entity_names, is_named, refs)

	def cluster(self, entities, is_named, refs):

		refs=self.name_cluster(entities, is_named, refs)
		clusters={}

		for i, val in enumerate(refs):
			ref=refs[i]
			if ref not in clusters:
				clusters[ref]=Counter()
			clusters[ref][' '.join(entities[i])]+=1

		# if two clusters have significant overlap in mention phrases, merge them into one
		for ref in clusters:
			for ref2 in clusters:
				if ref == ref2 or clusters[ref] is None or clusters[ref2] is None or ref == -1 or ref2 == -1 or ref == 0 or ref2 == 0:
					continue

				# find which cluster is bigger and should contain the other
				sum1=sum(clusters[ref].values())
				sum2=sum(clusters[ref2].values())

				big=ref
				small=ref2
				if sum2 > sum1:
					big=ref2
					small=ref

				sim=self.calc_overlap(clusters[small], clusters[big])
				if sim > 0.9:

					for k,v in clusters[small].most_common():
						clusters[big][k]+=v

					for idx, r in enumerate(refs):
						if r == small:
							refs[idx]=big

					clusters[small]=None

		counts=Counter()
		for ref in clusters:
			if clusters[ref] is not None:
				counts[ref]=sum(clusters[ref].values())

		return refs

if __name__ == "__main__":

	aliasFile = pkg_resources.resource_filename(__name__, "data/aliases.txt")
	resolver=NameCoref(aliasFile)
	resolver.process(sys.argv[1])




import sys
import argparse
from transformers import logging
logging.set_verbosity_error()
from pathlib import Path
import os
from booknlp.english.english_booknlp import EnglishBookNLP

class BookNLP():

	def __init__(self, language, model_params):

		if language == "en":
			self.booknlp=EnglishBookNLP(model_params)

	def process(self, inputFile, outputFolder, idd):
		self.booknlp.process(inputFile, outputFolder, idd)


def proc():

	parser = argparse.ArgumentParser()
	parser.add_argument('-l','--language', help='Currently on {en}', required=True)
	parser.add_argument('-i','--inputFile', help='Filename to run BookNLP on', required=True)
	parser.add_argument('-o','--outputFolder', help='Folder to write results to', required=True)
	parser.add_argument('--id', help='ID of text (for creating filenames within output folder)', required=True)

	args = vars(parser.parse_args())

	language=args["language"]
	inputFile=args["inputFile"]
	outputFolder=args["outputFolder"]
	idd=args["id"]

	print("tagging %s" % inputFile)
	
	valid_languages=set(["en"])
	if language not in valid_languages:
		print("%s not recognized; supported languages: %s" % (language, valid_languages))
		sys.exit(1)


	model_params={
		"pipeline":"entity,quote,supersense,event,coref", "model":"small", 
	}

	booknlp=BookNLP(language, model_params)
	booknlp.process(inputFile, outputFolder, idd)
		
if __name__ is "__main__":
	proc()


import re
from collections import Counter

class QuoteTagger:
	

	def tag(self, toks):

		predictions=[]
		currentQuote=[]
		curStartTok=None
		lastPar=None

		quote_symbols=Counter()

		for tok in toks:
			if tok.text == "“" or tok.text == "”" or tok.text == "\"" or tok.text == "“":
				quote_symbols["DOUBLE_QUOTE"]+=1
			elif tok.text == "‘" or tok.text == "’" or tok.text == "'":
				quote_symbols["SINGLE_QUOTE"]+=1
			elif tok.text == "—":
				quote_symbols["DASH"]+=1


		quote_symbol="DOUBLE_QUOTE"
		if len(quote_symbols) > 0:
			quote_symbol=quote_symbols.most_common()[0][0]

		for tok in toks:

			w=tok.text

			for w_idx, w_char in enumerate(w):
				if w_char== "“" or w_char == "”" or w_char == "\"":
					w="DOUBLE_QUOTE"
				elif w_char == "‘" or w_char == "’" or w_char == "'":
					if w_idx == 0:
						suff=w[w_idx+1:]
						if suff != "s" and suff != "d" and suff != "ll" and suff != "ve":
							w="SINGLE_QUOTE"

			# start over at each new paragraph
			if tok.paragraph_id != lastPar and lastPar is not None:

				if len(currentQuote) > 0:
					predictions.append((curStartTok, tok.token_id-1))
				curStartTok=None
				currentQuote=[]

			if w == quote_symbol:

				if curStartTok is not None:

					if len(currentQuote) > 0:
						predictions.append((curStartTok, tok.token_id))
						currentQuote.append(tok.text)

					curStartTok=None
					currentQuote=[]
				else:
					curStartTok=tok.token_id

			
			if curStartTok is not None:
				currentQuote.append(tok.text)

			lastPar=tok.paragraph_id

		for start, end in predictions:
			for i in range(start, end+1):
				toks[i].inQuote=True

		return predictions






import sys

def b3(gold, system):

	goldClusters={}
	for key in gold: 
		entity_id=gold[key]
		if entity_id not in goldClusters:
			goldClusters[entity_id]=[]
		goldClusters[entity_id].append(key)

	systemClusters={}
	for key in system: 
		entity_id=system[key]
		if entity_id not in systemClusters:
			systemClusters[entity_id]=[]
		systemClusters[entity_id].append(key)

	precision=0.
	recall=0.

	bigP=0.
	bigR=0.
	n=0.

	for mention in gold:

		goldCluster=set(goldClusters[gold[mention]])
		systemCluster=set(systemClusters[system[mention]])
		intersection=float(len(goldCluster.intersection(systemCluster)))

		precision=intersection/len(systemCluster)
		recall=intersection/len(goldCluster)

		bigP+=precision
		bigR+=recall

		n+=1

	precision=bigP/n if n > 0 else 0
	recall=bigR/n if n > 0 else 0
	F=(2*precision*recall)/(precision+recall) if (precision+recall) > 0 else 0


	return precision, recall, F


def read_data(filename):
	gold={}
	system={}

	with open(filename) as file:
		for idx, line in enumerate(file):
			cols=line.rstrip().split("\t")
			gold[idx]=cols[1]
			system[idx]=cols[2]

	return gold, system

if __name__ == "__main__":
	gold, system=read_data(sys.argv[1])
	precision, recall, F=b3(gold, system)
	print("F1: %.3f\tP: %.3f, R: %.3f" % (F, precision, recall))





import sys
import spacy
import copy
from booknlp.common.pipelines import SpacyPipeline
from booknlp.english.entity_tagger import LitBankEntityTagger
from booknlp.english.gender_inference_model_1 import GenderEM
from booknlp.english.name_coref import NameCoref
from booknlp.english.litbank_coref import LitBankCoref
from booknlp.english.litbank_quote import QuoteTagger
from booknlp.english.bert_qa import QuotationAttribution
from os.path import join
import os
import json
from collections import Counter
from html import escape
import time
from pathlib import Path
import urllib.request 
import pkg_resources
import torch

class EnglishBookNLP:

	def __init__(self, model_params):

		with torch.no_grad():

			start_time = time.time()

			print(model_params)

			spacy_model="en_core_web_sm"
			if "spacy_model" in model_params:
				spacy_model=model_params["spacy_model"]

			spacy_nlp = spacy.load(spacy_model, disable=["ner"])

			valid_keys=set("entity,event,supersense,quote,coref".split(","))
			
			pipes=model_params["pipeline"].split(",")

			self.gender_cats= [ ["he", "him", "his"], ["she", "her"], ["they", "them", "their"], ["xe", "xem", "xyr", "xir"], ["ze", "zem", "zir", "hir"] ] 

			if "referential_gender_cats" in model_params:
				self.gender_cats=model_params["referential_gender_cats"]

			home = str(Path.home())
			modelPath=os.path.join(home, "booknlp_models")
			if "model_path"  in model_params:			
				modelPath=model_params["model_path"]

			if not Path(modelPath).is_dir():
				Path(modelPath).mkdir(parents=True, exist_ok=True)

			if model_params["model"] == "big":
				entityName="entities_google_bert_uncased_L-6_H-768_A-12-v1.0.model"
				corefName="coref_google_bert_uncased_L-12_H-768_A-12-v1.0.model"
				quoteAttribName="speaker_google_bert_uncased_L-12_H-768_A-12-v1.0.1.model"

				self.entityPath=os.path.join(modelPath, entityName)
				if not Path(self.entityPath).is_file():
					print("downloading %s" % entityName)
					urllib.request.urlretrieve("http://ischool.berkeley.edu/~dbamman/booknlp_models/%s" % entityName, self.entityPath)

				self.coref_model=os.path.join(modelPath, corefName)
				if not Path(self.coref_model).is_file():
					print("downloading %s" % corefName)
					urllib.request.urlretrieve("http://ischool.berkeley.edu/~dbamman/booknlp_models/%s" % corefName, self.coref_model)

				self.quoteAttribModel=os.path.join(modelPath, quoteAttribName)
				if not Path(self.quoteAttribModel).is_file():
					print("downloading %s" % quoteAttribName)
					urllib.request.urlretrieve("http://ischool.berkeley.edu/~dbamman/booknlp_models/%s" % quoteAttribName, self.quoteAttribModel)


			elif model_params["model"] == "small":
				entityName="entities_google_bert_uncased_L-4_H-256_A-4-v1.0.model"
				corefName="coref_google_bert_uncased_L-2_H-256_A-4-v1.0.model"
				quoteAttribName="speaker_google_bert_uncased_L-8_H-256_A-4-v1.0.1.model"

				self.entityPath=os.path.join(modelPath, entityName)
				if not Path(self.entityPath).is_file():
					print("downloading %s" % entityName)
					urllib.request.urlretrieve("http://ischool.berkeley.edu/~dbamman/booknlp_models/%s" % entityName, self.entityPath)

				self.coref_model=os.path.join(modelPath, corefName)
				if not Path(self.coref_model).is_file():
					print("downloading %s" % corefName)
					urllib.request.urlretrieve("http://ischool.berkeley.edu/~dbamman/booknlp_models/%s" % corefName, self.coref_model)

				self.quoteAttribModel=os.path.join(modelPath, quoteAttribName)
				if not Path(self.quoteAttribModel).is_file():
					print("downloading %s" % quoteAttribName)
					urllib.request.urlretrieve("http://ischool.berkeley.edu/~dbamman/booknlp_models/%s" % quoteAttribName, self.quoteAttribModel)

			elif model_params["model"] == "custom":
				self.entityPath=model_params["entity_model_path"]
				self.coref_model=model_params["coref_model_path"]
				self.quoteAttribModel=model_params["quote_attribution_model_path"]


			self.doEntities=self.doCoref=self.doQuoteAttrib=self.doSS=self.doEvent=False

			for pipe in pipes:
				if pipe not in valid_keys:
					print("unknown pipe: %s" % pipe)
					sys.exit(1)
				if pipe == "entity":
					self.doEntities=True
				elif pipe == "event":
					self.doEvent=True
				elif pipe == "coref":
					self.doCoref=True
				elif pipe == "supersense":
					self.doSS=True
				elif pipe == "quote":
					self.doQuoteAttrib=True

			tagsetPath="data/entity_cat.tagset"
			tagsetPath = pkg_resources.resource_filename(__name__, tagsetPath)


			if "referential_gender_hyperparameterFile" in model_params:
				self.gender_hyperparameterFile=model_params["referential_gender_hyperparameterFile"]
			else:
				self.gender_hyperparameterFile = pkg_resources.resource_filename(__name__, "data/gutenberg_prop_gender_terms.txt")
			
			pronominalCorefOnly=True

			if "pronominalCorefOnly" in model_params:
				pronominalCorefOnly=model_params["pronominalCorefOnly"]

			if not self.doEntities and self.doCoref:
				print("coref requires entity tagging")
				sys.exit(1)

			if not self.doQuoteAttrib and self.doCoref:
				print("coref requires quotation attribution")
				sys.exit(1)
			if not self.doEntities and self.doQuoteAttrib:
				print("quotation attribution requires entity tagging")
				sys.exit(1)	


			self.quoteTagger=QuoteTagger()

			if self.doEntities:
				self.entityTagger=LitBankEntityTagger(self.entityPath, tagsetPath)
				aliasPath = pkg_resources.resource_filename(__name__, "data/aliases.txt")
				self.name_resolver=NameCoref(aliasPath)


			if self.doQuoteAttrib:
				self.quote_attrib=QuotationAttribution(self.quoteAttribModel)

			
			if self.doCoref:
				self.litbank_coref=LitBankCoref(self.coref_model, self.gender_cats, pronominalCorefOnly=pronominalCorefOnly)

			self.tagger=SpacyPipeline(spacy_nlp)

			print("--- startup: %.3f seconds ---" % (time.time() - start_time))

	def get_syntax(self, tokens, entities, assignments, genders):

		def check_conj(tok, tokens):
			if tok.deprel == "conj" and tok.dephead != tok.token_id:
				# print("found conj", tok.text)
				return tokens[tok.dephead]
			return tok

		def get_head_in_range(start, end, tokens):
			for i in range(start, end+1):
				if tokens[i].dephead < start or tokens[i].dephead > end:
					return tokens[i]
			return None

		agents={}
		patients={}
		poss={}
		mods={}
		prop_mentions={}
		pron_mentions={}
		nom_mentions={}
		keys=Counter()


		toks_by_children={}
		for tok in tokens:
			if tok.dephead not in toks_by_children:
				toks_by_children[tok.dephead]={}
			toks_by_children[tok.dephead][tok]=1

		for idx, (start_token, end_token, cat, phrase) in enumerate(entities):
			ner_prop=cat.split("_")[0]
			ner_type=cat.split("_")[1]

			if ner_type != "PER":
				continue

			coref=assignments[idx]

			keys[coref]+=1
			if coref not in agents:
				agents[coref]=[]
				patients[coref]=[]
				poss[coref]=[]
				mods[coref]=[]
				prop_mentions[coref]=Counter()
				pron_mentions[coref]=Counter()
				nom_mentions[coref]=Counter()

			if ner_prop == "PROP":
				prop_mentions[coref][phrase]+=1
			elif ner_prop == "PRON":
				pron_mentions[coref][phrase]+=1
			elif ner_prop == "NOM":
				nom_mentions[coref][phrase]+=1


			tok=get_head_in_range(start_token, end_token, tokens)
			if tok is not None:

				tok=check_conj(tok, tokens)
				head=tokens[tok.dephead]

				# nsubj
				# mod
				if tok.deprel == "nsubj" and head.lemma == "be":
					for sibling in toks_by_children[head.token_id]:

						# "he was strong and happy", where happy -> conj -> strong -> attr/acomp -> be
						sibling_id=sibling.token_id
						sibling_tok=tokens[sibling_id]
						if (sibling_tok.deprel == "attr" or sibling_tok.deprel == "acomp") and (sibling_tok.pos == "NOUN" or sibling_tok.pos == "ADJ"):
							mods[coref].append({"w":sibling_tok.text, "i":sibling_tok.token_id})

							if sibling.token_id in toks_by_children:
								for grandsibling in toks_by_children[sibling.token_id]:
									grandsibling_id=grandsibling.token_id
									grandsibling_tok=tokens[grandsibling_id]

									if grandsibling_tok.deprel == "conj" and (grandsibling_tok.pos == "NOUN" or grandsibling_tok.pos == "ADJ"):
										mods[coref].append({"w":grandsibling_tok.text, "i":grandsibling_tok.token_id})



				# ("Bill and Ted ran" conj captured by check_conj above)
				elif tok.deprel == "nsubj" and head.pos == ("VERB"):
					agents[coref].append({"w":head.text, "i":head.token_id})

				# "Bill ducked and ran", where ran -> conj -> ducked
					for sibling in toks_by_children[head.token_id]:
						sibling_id=sibling.token_id
						sibling_tok=tokens[sibling_id]
						if sibling_tok.deprel == "conj" and sibling_tok.pos == "VERB":
							agents[coref].append({"w":sibling_tok.text, "i":sibling_tok.token_id})
				
				# "Jack was hit by John and William" conj captured by check_conj above
				elif tok.deprel == "pobj" and head.deprel == "agent":
					# not root
					if head.dephead != head.token_id:
						grandparent=tokens[head.dephead]
						if grandparent.pos.startswith("V"):
							agents[coref].append({"w":grandparent.text, "i":grandparent.token_id})


				# patient ("He loved Bill and Ted" conj captured by check_conj above)
				elif (tok.deprel == "dobj" or tok.deprel == "nsubjpass") and head.pos == "VERB":
					patients[coref].append({"w":head.text, "i":head.token_id})


				# poss

				elif tok.deprel == "poss":
					poss[coref].append({"w":head.text, "i":head.token_id})

					# "her house and car", where car -> conj -> house
					for sibling in toks_by_children[head.token_id]:
						sibling_id=sibling.token_id
						sibling_tok=tokens[sibling_id]
						if sibling_tok.deprel == "conj":
							poss[coref].append({"w":sibling_tok.text, "i":sibling_tok.token_id})
					

		data={}
		data["characters"]=[]

		for coref, total_count in keys.most_common():

			# must observe a character at least *twice*

			if total_count > 1:
				chardata={}
				chardata["agent"]=agents[coref]
				chardata["patient"]=patients[coref]
				chardata["mod"]=mods[coref]
				chardata["poss"]=poss[coref]
				chardata["id"]=coref
				if coref in genders:
					chardata["g"]=genders[coref]
				else:
					chardata["g"]=None
				chardata["count"]=total_count

				mentions={}

				pnames=[]
				for k,v in prop_mentions[coref].most_common():
					pnames.append({"c":v, "n":k})
				mentions["proper"]=pnames

				nnames=[]
				for k,v in nom_mentions[coref].most_common():
					nnames.append({"c":v, "n":k})
				mentions["common"]=nnames

				prnames=[]
				for k,v in pron_mentions[coref].most_common():
					prnames.append({"c":v, "n":k})
				mentions["pronoun"]=prnames

				chardata["mentions"]=mentions

				
				data["characters"].append(chardata)
			
		return data
			


	def process(self, filename, outFolder, idd):		

		with torch.no_grad():

			start_time = time.time()
			originalTime=start_time

			with open(filename) as file:
				data=file.read()

				if len(data) == 0:
					print("Input file is empty: %s" % filename)
					return 

				try:
					os.makedirs(outFolder)
				except FileExistsError:
					pass

					
				tokens=self.tagger.tag(data)
				
				print("--- spacy: %.3f seconds ---" % (time.time() - start_time))
				start_time=time.time()

				if self.doEvent or self.doEntities or self.doSS:

					entity_vals=self.entityTagger.tag(tokens, doEvent=self.doEvent, doEntities=self.doEntities, doSS=self.doSS)
					entity_vals["entities"]=sorted(entity_vals["entities"])
					if self.doSS:
						supersense_entities=entity_vals["supersense"]
						with open(join(outFolder, "%s.supersense" % (idd)), "w", encoding="utf-8") as out:
							out.write("start_token\tend_token\tsupersense_category\ttext\n")
							for start, end, cat, text in supersense_entities:
								out.write("%s\t%s\t%s\t%s\n" % (start, end, cat, text))

					if self.doEvent:
						events=entity_vals["events"]
						for token in tokens:
							if token.token_id in events:
								token.event="EVENT"

					with open(join(outFolder, "%s.tokens" % (idd)), "w", encoding="utf-8") as out:
						out.write("%s\n" % '\t'.join(["paragraph_ID", "sentence_ID", "token_ID_within_sentence", "token_ID_within_document", "word", "lemma", "byte_onset", "byte_offset", "POS_tag", "fine_POS_tag", "dependency_relation", "syntactic_head_ID", "event"]))
						for token in tokens:
							out.write("%s\n" % token)

					print("--- entities: %.3f seconds ---" % (time.time() - start_time))
					start_time=time.time()

				in_quotes=[]
				quotes=self.quoteTagger.tag(tokens)

				print("--- quotes: %.3f seconds ---" % (time.time() - start_time))
				start_time=time.time()

				if self.doQuoteAttrib:

					entities=entity_vals["entities"]
					attributed_quotations=self.quote_attrib.tag(quotes, entities, tokens)

					print("--- attribution: %.3f seconds ---" % (time.time() - start_time))
					# return time.time() - start_time
					start_time=time.time()

				if self.doEntities:

					entities=entity_vals["entities"]
		
					in_quotes=[]

					for start, end, cat, text in entities:
	
						if tokens[start].inQuote or tokens[end].inQuote:
							in_quotes.append(1)
						else:
							in_quotes.append(0)


					# Create entity for first-person narrator, if present
					refs=self.name_resolver.cluster_narrator(entities, in_quotes, tokens)
				
					# Cluster non-PER PROP mentions that are identical
					refs=self.name_resolver.cluster_identical_propers(entities, refs)

					# Cluster mentions of named people
					refs=self.name_resolver.cluster_only_nouns(entities, refs, tokens)

					print("--- name coref: %.3f seconds ---" % (time.time() - start_time))

					start_time=time.time()

					# Infer referential gender from he/she/they mentions around characters
					
					genderEM=GenderEM(tokens=tokens, entities=entities, refs=refs, genders=self.gender_cats, hyperparameterFile=self.gender_hyperparameterFile)
					genders=genderEM.tag(entities, tokens, refs)
				
				assignments=None
				if self.doEntities:
					assignments=copy.deepcopy(refs)

				if self.doCoref:
					torch.cuda.empty_cache()
					assignments=self.litbank_coref.tag(tokens, entities, refs, genders, attributed_quotations, quotes)

					print("--- coref: %.3f seconds ---" % (time.time() - start_time))
					start_time=time.time()

					ent_names={}
					for a, e in zip(assignments, entities):
						if a not in ent_names:
							ent_names[a]=Counter()
						ent_names[a][e[3]]+=1
				
					# Update gender estimates from coref data
					genders=genderEM.update_gender_from_coref(genders, entities, assignments)

					chardata=self.get_syntax(tokens, entities, assignments, genders)
					with open(join(outFolder, "%s.book" % (idd)), "w", encoding="utf-8") as out:
						json.dump(chardata, out)

				if self.doEntities:
					# Write entities and coref			
					with open(join(outFolder, "%s.entities" % (idd)), "w", encoding="utf-8") as out:
						out.write("COREF\tstart_token\tend_token\tprop\tcat\ttext\n")
						for idx, assignment in enumerate(assignments):
							start, end, cat, text=entities[idx]
							ner_prop=cat.split("_")[0]
							ner_type=cat.split("_")[1]
							out.write("%s\t%s\t%s\t%s\t%s\t%s\n" % (assignment, start, end, ner_prop, ner_type, text))


				if self.doQuoteAttrib:
					with open(join(outFolder, "%s.quotes" % (idd)), "w", encoding="utf-8") as out:
						out.write('\t'.join(["quote_start", "quote_end", "mention_start", "mention_end", "mention_phrase", "char_id", "quote"]) + "\n")

						for idx, line in enumerate(attributed_quotations):
							q_start, q_end=quotes[idx]
							mention=attributed_quotations[idx]
							if mention is not None:
								entity=entities[mention]
								speaker_id=assignments[mention]
								e_start=entity[0]
								e_end=entity[1]
								cat=entity[3]
								speak=speaker_id
							else:
								e_start=None
								e_end=None
								cat=None
								speak=None
							quote=[tok.text for tok in tokens[q_start:q_end+1]]
							out.write("%s\t%s\t%s\t%s\t%s\t%s\t%s\n" % (q_start, q_end, e_start, e_end, cat, speak, ' '.join(quote)))
					
						out.close()

				if self.doQuoteAttrib and self.doCoref:

					# get canonical name for character
					names={}
					for idx, (start, end, cat, text) in enumerate(entities):
						coref=assignments[idx]
						if coref not in names:
							names[coref]=Counter()
						ner_prop=cat.split("_")[0]
						ner_type=cat.split("_")[1]
						if ner_prop == "PROP":
							names[coref][text.lower()]+=10
						elif ner_prop == "NOM":
							names[coref][text.lower()]+=1
						else:
							names[coref][text.lower()]+=.001


					with open(join(outFolder, "%s.book.html" % (idd)), "w", encoding="utf-8") as out:
						out.write("<html>")
						out.write("""<head>
		  <meta charset="UTF-8">
		</head>""")
						out.write("<h2>Named characters</h2>\n")
						for character in chardata["characters"]:
							char_id=character["id"]

							proper_names=character["mentions"]["proper"]
							if len(proper_names) > 0 or char_id == 0: # 0=narrator
								proper_name_list="/".join(["%s (%s)" % (name["n"], name["c"]) for name in proper_names])

								common_names=character["mentions"]["common"]
								common_name_list="/".join(["%s (%s)" % (name["n"], name["c"]) for name in common_names])

								char_count=character["count"]

								if char_id == 0:
									if len(proper_name_list) == 0:
										proper_name_list="[NARRATOR]"
									else:
										proper_name_list+="/[NARRATOR]"
								out.write("%s %s %s <br />\n" % (char_count, proper_name_list, common_name_list))

				
						out.write("<p>\n")

						out.write("<h2>Major entities (proper, common)</h2>")

						major_places={}
						for prop in ["PROP", "NOM"]:
							major_places[prop]={}
							for cat in ["FAC", "GPE", "LOC", "PER", "ORG", "VEH"]:
								major_places[prop][cat]={}

						for idx, (start, end, cat, text) in enumerate(entities):
							coref=assignments[idx]
			
							ner_prop=cat.split("_")[0]
							ner_type=cat.split("_")[1]
							if ner_prop != "PRON":
								if coref not in major_places[ner_prop][ner_type]:
									major_places[ner_prop][ner_type][coref]=Counter()
								major_places[ner_prop][ner_type][coref][text]+=1

						max_entities_to_display=10
						for cat in ["FAC", "GPE", "LOC", "PER", "ORG", "VEH"]:
							out.write("<h3>%s</h3>" % cat)
							for prop in ["PROP", "NOM"]:
								freqs={}
								for coref in major_places[prop][cat]:
									freqs[coref]=sum(major_places[prop][cat][coref].values())

								sorted_freqs=sorted(freqs.items(), key=lambda x: x[1], reverse=True)
								for k,v in sorted_freqs[:max_entities_to_display]:
									ent_names=[]
									for name, count in major_places[prop][cat][k].most_common():
										ent_names.append("%s" % (name))
									out.write("%s %s <br />"% (v, '/'.join(ent_names)))
								out.write("<p>")



						out.write("<h2>Text</h2>\n")
						

						beforeToks=[""]*len(tokens)
						afterToks=[""]*len(tokens)

						lastP=None

						for idx, (start, end, cat, text) in enumerate(entities):
							coref=assignments[idx]
							name=names[coref].most_common(1)[0][0]
							beforeToks[start]+="<font color=\"#D0D0D0\">[</font>"
							afterToks[end]="<font color=\"#D0D0D0\">]</font><font color=\"#FF00FF\"><sub>%s-%s</sub></font>" % (coref, name) + afterToks[end]

						for idx, (start, end) in enumerate(quotes):
							mention_id=attributed_quotations[idx]
							if mention_id is not None:
								speaker_id=assignments[mention_id]
								name=names[speaker_id].most_common(1)[0][0]
							else:
								speaker_id="None"
								name="None"
							beforeToks[start]+="<font color=\"#666699\">"
							afterToks[end]+="</font><sub>[%s-%s]</sub>" % (speaker_id, name)

						for idx in range(len(tokens)):
							if tokens[idx].paragraph_id != lastP:
								out.write("<p />")
							out.write("%s%s%s " % (beforeToks[idx], escape(tokens[idx].text), afterToks[idx])) 
							lastP=tokens[idx].paragraph_id	

						
						out.write("</html>")

				print("--- TOTAL (excl. startup): %.3f seconds ---, %s words" % (time.time() - originalTime, len(tokens)))
				return time.time() - originalTime





import os

def read_tagset(filename):
	tags={}
	with open(filename) as file:
		for line in file:
			cols=line.rstrip().split("\t")
			tags[cols[0]]=int(cols[1])
	return tags


def read_filenames(filename):
	inpaths=[]
	outpaths=[]
	with open(filename) as file:
		for line in file:
			cols=line.rstrip().split("\t")
			
			if len(cols) == 2:
				inpaths.append(cols[0])
				outpaths.append(cols[1])

	return inpaths, outpaths


def read_booknlp(path, model):

	sentences=[]
	original_sentences=[]

	sentence=[]
	sentence.append(["[CLS]"])
	
	orig_sentence=[]
	orig_sentence.append(("[CLS]", -1))
	length=0
	
	max_sentence_length=500


	with open(path, encoding="utf-8") as file:
		header=file.readline().split("\t")

		s_idx=header.index("sentenceID")
		t_idx=header.index("tokenId")
		w_idx=header.index("originalWord")
		
		lastSentence=None

		for line in file:
			cols=line.rstrip().split("\t")
			s_id=cols[s_idx]
			t_id=cols[t_idx]
			w=cols[w_idx]

			toks=model.tokenizer.tokenize(w)
			
			if s_id != lastSentence or length + len(toks) > max_sentence_length:
				if len(sentence) > 0:
					sentence.append(["[SEP]"])
					sentences.append(sentence)

					orig_sentence.append((-1, "[SEP]"))
					original_sentences.append(orig_sentence)

					sentence=[]
					sentence.append(["[CLS]"])

					orig_sentence=[]
					orig_sentence.append((-1, "[CLS]"))

					length=0

			length+=len(toks)

			sentence.append([w])
			orig_sentence.append((t_id, w))

			lastSentence=s_id

	if len(sentence) > 1:
		sentence.append(["[SEP]"])
		sentences.append(sentence)

		orig_sentence.append((-1, "[SEP]"))
		original_sentences.append(orig_sentence)

	return sentences, original_sentences


def read_annotations(filename, tagset, labeled):

	""" Read tsv data and return sentences and [word, tag, sentenceID, filename] list """

	with open(filename, encoding="utf-8") as f:
		sentence = []
		sentence.append(["[CLS]", -100, -100, -100, -100, -100, -100 -1, -1, None])
		sentences = []
		sentenceID=0
		for line in f:
			if len(line) > 0:
				if line == '\n':
					sentenceID+=1

					sentence.append(["[SEP]", -100, -100, -100, -100, -100, -100 -1, -1, None])

					if len(sentence) > 2:
						sentences.append(sentence)

					sentence = []
					sentence.append(["[CLS]", -100, -100, -100, -100, -100, -100 -1, -1, None])


				else:
					data=[]
					split_line = line.rstrip().split('\t')

					data.append(split_line[0])
					data.append(tagset[split_line[1]] if labeled else 0)
					data.append(tagset[split_line[2]] if labeled else 0)
					data.append(tagset[split_line[3]] if labeled else 0)
					data.append(tagset[split_line[4]] if labeled else 0)
					data.append(tagset[split_line[5]] if labeled else 0)

					data.append(sentenceID)
					data.append(filename)

					sentence.append(data)
		
		sentence.append(["[SEP]", -100, -100, -100, -100, -100, -100 -1, -1, None])
		if len(sentence) > 2:
			sentences.append(sentence)

	return sentences

def prepare_annotations_from_file(filename, tagset, labeled=True):

	""" Read a single file of annotations, returning:
		-- a list of sentences, each a list of [word, label, sentenceID, filename]
	"""

	sentences = []
	annotations = read_annotations(filename, tagset, labeled)
	sentences += annotations
	return sentences

def prepare_annotations_from_folder(folder, tagset, labeled=True):

	""" Read folder of annotations, returning:
		-- a list of sentences, each a list of [word, label, sentenceID, filename]
	"""

	sentences = []
	for filename in os.listdir(folder):
		print(filename)
		annotations = read_annotations(os.path.join(folder,filename), tagset, labeled)
		sentences += annotations
	print("num sentences: %s" % len(sentences))
	return sentences



import subprocess, re, sys

def get_coref_score(metric, path_to_scorer, gold=None, preds=None):

	output=subprocess.check_output(["perl", path_to_scorer, metric, preds, gold]).decode("utf-8")
	output=output.split("\n")[-3]
	matcher=re.search("Coreference: Recall: \(.*?\) (.*?)%	Precision: \(.*?\) (.*?)%	F1: (.*?)%", output)
	if matcher is not None:
		recall=float(matcher.group(1))
		precision=float(matcher.group(2))
		f1=float(matcher.group(3))
	return recall, precision, f1

def get_conll(path_to_scorer, gold=None, preds=None):
	bcub_r, bcub_p, bcub_f=get_coref_score("bcub", path_to_scorer, gold, preds)
	muc_r, muc_p, muc_f=get_coref_score("muc", path_to_scorer, gold, preds)
	ceaf_r, ceaf_p, ceaf_f=get_coref_score("ceafe", path_to_scorer, gold, preds)

	print("bcub:\t%.1f" % bcub_f)
	print("muc:\t%.1f" % muc_f)
	print("ceaf:\t%.1f" % ceaf_f)
	avg=(bcub_f + muc_f + ceaf_f)/3.
	print("Average F1: %.1f" % (avg))

	# Generate Latex table
	# print("%.1f&%.1f&%.1f&%.1f" % (bcub_f, muc_f, ceaf_f, avg))

	return bcub_f, avg

if __name__ == "__main__":

	goldFile=sys.argv[1]
	predFile=sys.argv[2]
	scorer=sys.argv[3]
	bcub_f, avg=get_conll(scorer, gold=goldFile, preds=predFile)
	

	


from booknlp.english.tagger import Tagger
import torch
import re
import booknlp.common.layered_reader as layered_reader
import booknlp.common.sequence_layered_reader as sequence_layered_reader
import pkg_resources

class LitBankEntityTagger:
	def __init__(self, model_file, model_tagset):

		device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
		self.tagset=sequence_layered_reader.read_tagset(model_tagset)
		supersenseTagset = pkg_resources.resource_filename(__name__, "data/supersense.tagset")

		self.supersense_tagset=sequence_layered_reader.read_tagset(supersenseTagset)
		base_model=re.sub("google_bert", "google/bert", model_file.split("/")[-1])
		base_model=re.sub(".model", "", base_model)

		self.model = Tagger(freeze_bert=False, base_model=base_model, tagset_flat={"EVENT":1, "O":1}, supersense_tagset=self.supersense_tagset, tagset=self.tagset, device=device)

		self.model.to(device)
		self.model.load_state_dict(torch.load(model_file, map_location=device))
		wnsFile = pkg_resources.resource_filename(__name__, "data/wordnet.first.sense")
		self.wns=self.read_wn(wnsFile)

	def read_wn(self, filename):
		wns={}
		with open(filename) as file:
			for line in file:
				cols=line.rstrip().split("\t")
				word=cols[0]
				pos=cols[1]
				wn=int(cols[2].split(" ")[0])
				wns["%s.%s" % (word, pos)]=wn
		return wns

	def get_wn(self, supersense_batched_sents):


		wn_batches=[]

		for idx, b_sent in enumerate(supersense_batched_sents):

			max_len=0
			for sent in b_sent:
				if sent is not None:
					if len(sent) > max_len:
						max_len=len(sent)

			wn_senses=[]

			for sent in b_sent:

				wn=[]
				if sent is None:
					continue

				for word in sent:

					if word is None:
						wn.append(0)
					else:

						text=word.text
						pos=word.pos
						if pos == "NOUN":
							pos="n"
						elif pos == "VERB":
							pos="v"
						term=text.split(" ")[-1].lower()
						key="%s.%s" % (term, pos)
						if key in self.wns:
							wn.append(self.wns[key])
						else:
							wn.append(1)

				for val in range(len(sent), max_len):
					wn.append(0)
				wn_senses.append(wn)

			wn_senses=torch.LongTensor(wn_senses)
			wn_batches.append(wn_senses)
		return wn_batches

	def tag(self, toks, doEvent=True, doEntities=True, doSS=True):

		max_sentence_length=500

		entities=[]
		supersense_entities=[]

		batch_size=32

		sents=[]
		o_sents=[]
		sent=[]
		o_sent=[]
		lastSid=None

		length=0

		for tok in toks:

			wptok=tok.text
			# working with uncased BERT models, so add a special tag to denote capitalization
			if wptok[0].lower() != wptok[0]:
				wptok="[CAP] " + wptok.lower()

			toks=self.model.tokenizer.tokenize(wptok)
			if lastSid is not None and (tok.sentence_id != lastSid or length + len(toks) > max_sentence_length):
				sents.append(sent)
				o_sents.append(o_sent)
				sent=[]
				o_sent=[]
				length=0
			
			sent.append(toks)
			o_sent.append(tok)

			lastSid=tok.sentence_id
			length+=len(toks)
		
		sents.append(sent)
		o_sents.append(o_sent)


		sentences=[]
		o_sentences=[]

		max_sentence_length=500

		sentence=[ ["[CLS]"] ]

		o_sent=[]

		cur_length=0

		for idx, sent in enumerate(sents):

			sent_len=0
			for toks in sent:
				sent_len+=len(toks)

			if sent_len + cur_length >= max_sentence_length:
				sentence.append(["[SEP]"])
				sentences.append(sentence)
				o_sentences.append(o_sent)
				sentence=[["[CLS]"]]
				o_sent=[]
				cur_length=0

			cur_length+=sent_len

			sentence.extend(sent)
			o_sent.extend(o_sents[idx])


		if len(sentence) > 1:		
			sentence.append(["[SEP]"])
			o_sentences.append(o_sent)
			sentences.append(sentence)

		sents=o_sentences

		batched_sents, batched_data, batched_mask, batched_transforms, batched_orig_token_lens, ordering, order_to_batch_map = layered_reader.get_batches(self.model, sentences, batch_size, self.tagset, training=False)
		
		batch_pos={}
		for idx, ind in enumerate(ordering):
			batch_id, batch_s, batch_position=order_to_batch_map[idx]
			if batch_id not in batch_pos:
				batch_pos[batch_id]=[None]*batch_s
			batch_pos[batch_id][batch_position]=[None]
			for tok in sents[ind]:
				# print(tok)
				batch_pos[batch_id][batch_position].append(tok)
			batch_pos[batch_id][batch_position].append(None)

		batched_pos=[None]*len(batch_pos)
		for i in range(len(batch_pos)):
			batched_pos[i]=batch_pos[i]

		wn_batches=self.get_wn(batched_pos)

		preds_in_order, events_in_order, supersense_preds_in_order=self.model.tag_all(wn_batches, batched_sents, batched_data, batched_mask, batched_transforms, batched_orig_token_lens, ordering, doEvent=doEvent, doEntities=doEntities, doSS=doSS)
		
		return_vals={}

		if doEntities:
			for idx, preds in enumerate(preds_in_order):
				for tmp, label, start, end in preds:
					start_token=sents[idx][start].token_id
					end_token=sents[idx][end-1].token_id
					phrase=' '.join([x.text for x in sents[idx][start:end]])
					phraseEndToken=int(end_token)
					if phraseEndToken == -2:
						phraseEndToken=start_token
					entities.append((start_token, phraseEndToken, label, phrase))
			return_vals["entities"]=entities

		if doSS:
			for idx, preds in enumerate(supersense_preds_in_order):
				for tmp, label, start, end in preds:
					start_token=sents[idx][start].token_id
					end_token=sents[idx][end-1].token_id
					phrase=' '.join([x.text for x in sents[idx][start:end]])
					phraseEndToken=int(end_token)
					if phraseEndToken == -2:
						phraseEndToken=start_token
					supersense_entities.append((start_token, phraseEndToken, label, phrase))
			return_vals["supersense"]=supersense_entities
			
		
		if doEvent:
			events={}
			for idx, preds in enumerate(events_in_order):

				for start in preds:
					start_token=sents[idx][start].token_id
					phrase=sents[idx][start].text
					events[start_token]=1
				return_vals["events"]=events

		return return_vals






import torch
import re
from booknlp.english.speaker_attribution import BERTSpeakerID
import numpy as np
import sys

PINK = '\033[95m'
ENDC = '\033[0m'

class QuotationAttribution:

	def __init__(self, modelFile):

		device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

		base_model=re.sub("google_bert", "google/bert", modelFile.split("/")[-1])
		base_model=re.sub(".model", "", base_model)

		self.model = BERTSpeakerID(base_model=base_model)
		self.model.load_state_dict(torch.load(modelFile, map_location=device))
		self.model.to(device)
		self.model.eval()

	def tag(self, quotes, entities, tokens):

		def get_base(start, end, preds):
			if (start, end) in preds:
				s,e=preds[(start,end)]
				return get_base(s,e,preds)

			return start, end

		attributions=[None]*len(quotes)

		entity_by_position={}
		for idx, (start, end, cat, text) in enumerate(entities):
			entity_by_position[start, end]=idx

		texts, metas, positions, global_entity_positions, quote_indexes=self.get_representation(quotes, entities, tokens)

		x_batches, m_batches, y_batches, o_batches=self.model.get_batches(texts, metas)

		all_preds={}
		quote_chain={}

		idd=0
		prediction_id=0

		for x1, m1, y1, o1 in zip(x_batches, m_batches, y_batches, o_batches):
			y_pred = self.model.forward(x1, m1)
			orig, meta=o1
			predictions=torch.argmax(y_pred, axis=1).detach().cpu().numpy()
			for idx, pred in enumerate(predictions):

				global_quote_id=quote_indexes[prediction_id]

				prediction=pred[0]
				quote_start, quote_end=quotes[global_quote_id]
				sent=orig[idx]
				predval=y1["eid"][idx][prediction]

				if prediction >= len(meta[idx][1]):
					prediction=torch.argmax(y_pred[idx][:len(meta[idx][1])])

				g_start, g_end=global_entity_positions[prediction_id][prediction]

				cat,start, end, orig_text=positions[prediction_id][prediction]

				if cat == "QUOTE":
					g_start, g_end=get_base(start, end, quote_chain)
	
				if (g_start, g_end) in entity_by_position:
					quote_chain[quote_start, quote_end]=g_start, g_end
					attributions[prediction_id]=entity_by_position[g_start, g_end]
				else:
					print("Cannot resolve quotation")

				ent_start, ent_end, lab, ent_eid=meta[idx][1][prediction]

				if ' '.join(sent[ent_start:ent_end]) == "[PAR]":
					print("Problem!!!! Linked [PAR]")
					sys.exit(1)
							
				prediction_id+=1

		return attributions



	def get_representation(self, quotes, entities, tokens, doLowerCase=True):

		def convert_word(word):
			if word == "[ALTQUOTE]" or word == "[PAR]" or word == "[QUOTE]":
				word=word

			elif doLowerCase:
				if word[0].lower() != word[0]:
					word="[CAP] " + word.lower()
				else:
					word=word.lower()

			return word

		window=50

		texts=[]
		metas=[]
		positions=[]

		global_positions=[]

		quote_indexes=[]

		end_quotes={}

		# mark the tokens that are in quotes and index quotes by their final token
		in_quotes=np.zeros(len(tokens))
		for idx, (q_start, q_end) in enumerate(quotes):
			end_quotes[q_end]=idx
			for k in range(q_start, q_end+1):
				in_quotes[k]=1

		entities_by_start={}
		for start, end, cat, text in entities:
			ner_prop=cat.split("_")[0]
			ner_type=cat.split("_")[1]

			if ner_type != "PER":
				continue

			if start not in entities_by_start:
				entities_by_start[start]={}
			entities_by_start[start][end]=1

		for q_id, (start_tok, end_tok) in enumerate(quotes):

			start=end_tok
			count=0
			wp_tok_count=0
			# go back *window* tokens, not counting tokens that are in quotes
			lastPar=None
			while start >= 0 and count < window and wp_tok_count + len(self.model.tokenizer.tokenize(convert_word(tokens[start].text))) < 350:
				if in_quotes[start] == 0:
					count+=1
					wp_tok_count+=len(self.model.tokenizer.tokenize(convert_word(tokens[start].text)))
				if start in end_quotes:
					wp_tok_count+=1
				if tokens[start].paragraph_id != lastPar:
					wp_tok_count+=1
				lastPar=tokens[start].paragraph_id

				start-=1

			start+=1

			# go ahead *window* tokens, not counting tokens that are in quotes
			count=0
			end=end_tok
			while end < len(tokens) and count < window and wp_tok_count + len(self.model.tokenizer.tokenize(convert_word(tokens[end].text))) < 475:
				if in_quotes[end] == 0:
					count+=1
					wp_tok_count+=len(self.model.tokenizer.tokenize(convert_word(tokens[end].text)))
				if end in end_quotes:
					wp_tok_count+=1
				if tokens[end].paragraph_id != lastPar:
					wp_tok_count+=1
				lastPar=tokens[end].paragraph_id

				end+=1

			end-=1

			if end < end_tok+1:
				end=end_tok+1


			toks=[]
			cands=[]

			lastPar=None

			inserts=np.zeros(end-start, dtype=int)

			# the offset keeps track of the difference between the original token position and the resulting token position
			# (after the addition of [PAR], [QUOTE], [ALTQUOTE] pseudo-tokens, and the subtraction of tokens within quotations)
			offset=0
			
			# reverse map maps the resulting tokens (which include [PAR], [QUOTE] etc.) to the original position
			reverse_map=[]
			altquote_map={}

			for i in range(start, end):

				tok=tokens[i]

				# if we cross a sentence boundary, add a [PAR] pseudo-token and adjust the offset to keep track of these pseudo-tokens
				if tok.paragraph_id != lastPar and lastPar is not None:
					toks.append("[PAR]")
					reverse_map.append(i)
					offset+=1

				# if the token is not in a quotation, add it to the context representation
				# if it is in a quotation, decrease the offset
				if not in_quotes[i]:
					toks.append(tokens[i].text)
					reverse_map.append(i)
				else:
					offset-=1

				# if the token marks the end of the target quotation, add a [QUOTE] pseudo-token to represent the quotation in its entirety
				# (note we don't include any words from within the quotation itself)
				if i == end_tok:
					toks.append("[QUOTE]")
					reverse_map.append(i)
					offset+=1

				# if the token ends a *different* quote, add an [ALTQUOTE] pseudo-token to represent it
				elif i in end_quotes:
					altquote_map[len(toks)]=quotes[end_quotes[i]]
					toks.append("[ALTQUOTE]")
					reverse_map.append(i)
					offset+=1

					(q_start, q_end)=quotes[end_quotes[i]]
		
					# if the alt quote occurs *before* the target quote, add it as an attribution candidate
					if q_end < end_tok:
						quotepos=i+inserts[i-start-1]-start
						cands.append((min(abs(q_end-start_tok), abs(q_start-end_tok)), quotepos, quotepos, None, "QUOTE", None, None))

				# inserts keeps track of the difference between the original token position and final one
				inserts[i-start]=offset

				lastPar=tok.paragraph_id

			for entity_start in range(start, end):

				if entity_start in entities_by_start:
					# only allow entities not in quotes to be speaker candidates
					if in_quotes[entity_start] == 0:
						for entity_end in entities_by_start[entity_start]:
							# the candidate entity needs to be completely contained within the context
							if entity_end < end and entity_end+inserts[entity_end-start]-start < len(toks):
								# cand tuple values:
								# distance between the entity and the quotation (in original tokens)
								# entity start position (in final tokens)
								# entity end position (in final tokens)
								# entity ID (here unknown)
								# whether the candidate is an entity ("ENT") or alt quote ("QUOTE")
								# entity start position (in original tokens)
								# entity end position (in original tokens)
								# print(entity_start, start, entity_end, len(inserts))
								# print(inserts[entity_start-start])
								# print(inserts[entity_end-start])
								cands.append((min(abs(entity_end-start_tok), abs(entity_start-end_tok)), entity_start+inserts[entity_start-start]-start, entity_end+inserts[entity_end-start]-start, None, "ENT", entity_start, entity_end))

			tot_toks=0
			for tok in toks:
				tot_toks+=len(self.model.tokenizer.tokenize(convert_word(tok)))
			if tot_toks > 500:
				raise ValueError("Quotation window is unexpectedly long: %s" % tot_toks)

			labels=[]

			abs_positions=[]
			g_positions=[]

			if len(cands) > 0:
				# sort the candidates by their distance to the target quote and consider only the 10 closest
				cands=sorted(cands)
				for dist, s, e, eid, cat, global_start, global_end in cands[:10]:
					
					if ' '.join(toks[s:e+1]) == "[PAR]":
						# skip
						continue

					labels.append((int(s),int(e+1),0, eid))
					adjusted_s=reverse_map[s]
					adjusted_e=reverse_map[e]

					g_positions.append((global_start, global_end))

					if ' '.join(toks[s:e+1]) != "[ALTQUOTE]":
					# 	if ' '.join(toks[s:e+1])  != "[PAR]":
					# 		if ' '.join(toks[s:e+1]) != ' '.join([t.text for t in tokens[global_start:global_end+1]]):
					# 			print("ASSERTION\t#%s#%s#" % (' '.join(toks[s:e+1]), ' '.join([t.text for t in tokens[global_start:global_end+1]])))

						abs_positions.append(("ENT", global_start, global_end+1, ' '.join([t.text for t in tokens[global_start:global_end+1]])))
					else:
						abs_positions.append(("QUOTE", altquote_map[s][0], altquote_map[s][1], ' '.join([t.text for t in tokens[altquote_map[s][0]:altquote_map[s][1]]])))

				index=toks.index("[QUOTE]")
				texts.append(toks)
				metas.append((None, labels, index))
				positions.append(abs_positions)
				global_positions.append(g_positions)
				quote_indexes.append(q_id)

		return texts, metas, positions, global_positions, quote_indexes
	



import sys
import numpy as np

def get_accuracy(golds, preds, tagset):
	cor=0.
	tot=0.
	for i in range(len(golds)):
		if golds[i] == preds[i]:
			cor+=1
		tot+=1

	acc=cor/tot
	print ("Accuracy: %.3f" % acc)

	return acc


def check_span_f1_two_dicts_subcat(gold, pred):


	golds={}
	preds={}

	for target_lab in ["PRON", "NOM", "PROP"]:
		golds[target_lab]={}
		preds[target_lab]={}
		
		for doc, lab, start, end in gold:
			# print(lab)
			subcat=lab.split("_")[0]
			if subcat == target_lab:
				golds[target_lab][(doc, lab, start, end)]=1

		for doc, lab, start, end in pred:
			subcat=lab.split("_")[0]
			if subcat == target_lab:
				preds[target_lab][(doc, lab, start, end)]=1
			
	cor=0.
	for g in gold:
		if g in pred:
			cor+=1

	precision=0
	if len(pred) > 0:
		precision=cor/len(pred)
	recall=0
	if len(gold) > 0:
		recall=cor/len(gold)
	mainF=0
	if (precision + recall) > 0:
		mainF=(2*precision*recall)/(precision+recall)

	print ("precision: %.3f %s/%s" % (precision, cor, len(pred)))
	print ("recall: %.3f %s/%s" % (recall, cor, len(gold)))
	print ("F: %.3f" % mainF)


	for target_lab in ["PRON", "NOM", "PROP"]:
		cor=0.
		for g in golds[target_lab]:
			if g in preds[target_lab]:
				cor+=1

		precision=0
		if len(preds[target_lab]) > 0:
			precision=cor/len(preds[target_lab])
		recall=0
		if len(golds[target_lab]) > 0:
			recall=cor/len(golds[target_lab])
		F=0
		if (precision + recall) > 0:
			F=(2*precision*recall)/(precision+recall)

		print ("\n\t%s precision: %.3f %s/%s" % (target_lab, precision, cor, len(preds[target_lab])))
		print ("\t%s recall: %.3f %s/%s" % (target_lab, recall, cor, len(golds[target_lab])))
		print ("\t%s F: %.3f" % (target_lab, F))


	return mainF


def check_span_f1_two_dicts(gold, pred):

	cor=0.
	for g in gold:
		if g in pred:
			cor+=1

	precision=0
	if len(pred) > 0:
		precision=cor/len(pred)
	recall=0
	if len(gold) > 0:
		recall=cor/len(gold)
	F=0
	if (precision + recall) > 0:
		F=(2*precision*recall)/(precision+recall)

	print ("precision: %.3f %s/%s" % (precision, cor, len(pred)))
	print ("recall: %.3f %s/%s" % (recall, cor, len(gold)))
	print ("F: %.3f" % F)

	return F

def check_span_f1_two_lists(gold, preds, orig_tagset):

	tagset={orig_tagset[v]:v for v in orig_tagset}
	print(tagset)
	start_idx=-1
	start_tag=None

	gold_spans={}

	for i, tag_idx in enumerate(gold):
		tag=tagset[int(tag_idx)]
		if tag == "O" or tag.startswith("B-"):
			if start_idx != -1:
				end_idx=i-1
				gold_spans[(start_idx, end_idx, start_tag)]=1
				start_idx=-1

		if tag.startswith("B-"):
			start_idx=i
			start_tag=tag.split("-")[-1]

	if start_idx != -1:
		end_idx=len(gold)
		gold_spans[(start_idx, end_idx, start_tag)]=1


	start_idx=-1
	start_tag=None

	pred_spans={}

	for i, tag_idx in enumerate(preds):
		tag=tagset[int(tag_idx)]
		if tag == "O" or tag.startswith("B-"):
			if start_idx != -1:
				end_idx=i-1
				pred_spans[(start_idx, end_idx, start_tag)]=1
				start_idx=-1

		if tag.startswith("B-"):
			start_idx=i
			start_tag=tag.split("-")[-1]

	if start_idx != -1:
		end_idx=len(gold)
		pred_spans[(start_idx, end_idx, start_tag)]=1

	correct=0.
	for tag in gold_spans:
		if tag in pred_spans:
			correct+=1


	trials=len(pred_spans)
	trues=len(gold_spans)

	p=0
	if trials > 0:
		p=correct/trials
	r=0
	if trues > 0:
		r=correct/trues

	f=0
	if (p+r) > 0:
		f=(2*p*r)/(p+r)

	print ("precision: %.3f %s/%s" % (p, correct, trials))
	print ("recall: %.3f %s/%s" % (r, correct, trues))
	print ("F: %.3f" % f)

	return f

def check_f1_two_lists(gold, preds, tagset):

	correct=0.
	trials=0.
	trues=0.
	for j in range(len(preds)):
		if preds[j] == 1:
			trials+=1
		if gold[j] == 1:
			trues+=1
		if preds[j] == gold[j] and preds[j] == 1:
			correct+=1

	p=0
	if trials > 0:
		p=correct/trials
	r=0
	if trues > 0:
		r=correct/trues

	f=0
	if (p+r) > 0:
		f=(2*p*r)/(p+r)

	print ("precision: %.3f %s/%s" % (p, correct, trials))
	print ("recall: %.3f %s/%s" % (r, correct, trues))
	print ("F: %.3f" % f)

	return f


def check_f1(data):
	correct=0.
	trials=0.
	trues=0.

	for sentence in data:
		for word in sentence:
			truth=word[0]
			pred=word[1]

			if pred == 1:
				trials+=1
			if truth == 1:
				trues+=1
			if pred == truth and pred == 1:
				correct+=1

	p=0.
	if trials > 0:
		p=correct/trials
	r=0.
	if trues > 0:
		r=correct/trues

	f=0.
	if (p+r) > 0:
		f=(2*p*r)/(p+r)

	return f, p, r, correct, trials, trues





import sys
import re
import random
from random import shuffle
from math import sqrt, exp, isnan
from transformers import BertTokenizer, BertModel
import torch.nn as nn
import torch
import numpy as np
import argparse
import json
from booknlp.common.b3 import b3

from collections import Counter

PINK = '\033[95m'
ENDC = '\033[0m'

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class BERTSpeakerID(nn.Module):

	def __init__(self, base_model=None):
		super().__init__()

		modelName=base_model
		modelName=re.sub("^speaker_", "", modelName)
		modelName=re.sub("-v\d.*$", "", modelName)
		matcher=re.search(".*-(\d+)_H-(\d+)_A-.*", modelName)
		bert_dim=0
		modelSize=0
		self.num_layers=0
		if matcher is not None:
			bert_dim=int(matcher.group(2))
			self.num_layers=min(4, int(matcher.group(1)))

			modelSize=self.num_layers*bert_dim

		assert bert_dim != 0

		self.tokenizer = BertTokenizer.from_pretrained(modelName, do_lower_case=False, do_basic_tokenize=False)
		self.tokenizer.add_tokens(["[QUOTE]", "[ALTQUOTE]", "[PAR]", "[CAP]"], special_tokens=True)
		self.bert = BertModel.from_pretrained(modelName)
		self.bert.resize_token_embeddings(len(self.tokenizer))
			
		self.tanh = nn.Tanh()
		self.fc = nn.Linear(2*bert_dim, 100)
		self.fc2 = nn.Linear(100, 1)

	def get_wp_position_for_all_tokens(self, words, doLowerCase=True):

		wps=[]

		# start with 1 for the inital [CLS] token
		cur=1
		for idx, word in enumerate(words):
			if word == "[ALTQUOTE]" or word == "[PAR]" or word == "[QUOTE]":
				word=word

			elif doLowerCase:
				if word[0].lower() != word[0]:
					word="[CAP] " + word.lower()
				else:
					word=word.lower()

			target=self.tokenizer.tokenize(word)
			wps.append((cur, cur+len(target)))
			cur+=len(target)
		
		return wps


	def get_batches(self, all_x, all_m, batch_size=32, doLowerCase=True):
				
		batches_o=[]	
		batches_x=[]
		batches_y=[]
		batches_m=[]
			
		for i in range(0, len(all_x), batch_size):
			
			current_batch_input_ids=[]
			current_batch_attention_mask=[]
			current_batch_matrix_cands=[]
			current_batch_matrix_quote=[]
			current_batch_y=[]
			current_batch_eid=[]
			current_quote_eids=[]

			xb=all_x[i:i+batch_size]
			mb=all_m[i:i+batch_size]
			for s, sent in enumerate(xb):

				sent_wp_tokens=[self.tokenizer.convert_tokens_to_ids("[CLS]")]
				attention_mask=[1]

				for word in sent:
					if word == "[ALTQUOTE]" or word == "[PAR]" or word == "[QUOTE]":
						word=word

					elif doLowerCase:
						if word[0].lower() != word[0]:
							word="[CAP] " + word.lower()
						else:
							word=word.lower()

					toks = self.tokenizer.tokenize(word)
					toks = self.tokenizer.convert_tokens_to_ids(toks)
					sent_wp_tokens.extend(toks)
					attention_mask.extend([1]*len(toks))
				sent_wp_tokens.append(self.tokenizer.convert_tokens_to_ids("[SEP]"))
				attention_mask.append(1)

				current_batch_input_ids.append(sent_wp_tokens)
				current_batch_attention_mask.append(attention_mask)

			max_len = max([len(s) for s in current_batch_input_ids])

			for j, sent in enumerate(current_batch_input_ids):
				for k in range(len(current_batch_input_ids[j]), max_len):
					current_batch_input_ids[j].append(0)
					current_batch_attention_mask[j].append(0)

			for j, (eid, cands, quote) in enumerate(mb):


				wps_all=self.get_wp_position_for_all_tokens(xb[j])

				current_quote_eids.append(eid)

				e1_start_wp, e1_end_wp=wps_all[quote]
				matrix_cands=np.zeros((10,max_len))
				matrix_quote=np.zeros((10,max_len))

				for l in range(10):
					for k in range(e1_start_wp, e1_end_wp):
						matrix_quote[l][k]=1./(e1_end_wp-e1_start_wp)

				y=[]
				eids=[]
				for c_idx, (start, end, truth, cand_eid) in enumerate(cands):

					e2_start_wp, _=wps_all[start]
					_, e2_end_wp=wps_all[end-1]

					for k in range(e2_start_wp, e2_end_wp):
						matrix_cands[c_idx][k]=1./(e2_end_wp-e2_start_wp)

					y.append(truth)
					eids.append(cand_eid)

				for l in range(len(y), 10):
					y.append(0)
					eids.append(None)

				current_batch_matrix_cands.append(matrix_cands)
				current_batch_matrix_quote.append(matrix_quote)
				current_batch_y.append(y)
				current_batch_eid.append(eids)


			batches_o.append((xb, mb))
			batches_x.append({"toks": torch.LongTensor(current_batch_input_ids).to(device), "mask":torch.LongTensor(current_batch_attention_mask).to(device)})
			batches_m.append({"cands":torch.FloatTensor(np.array(current_batch_matrix_cands)).to(device), "quote":torch.FloatTensor(np.array(current_batch_matrix_quote)).to(device)})
			batches_y.append({"y":torch.LongTensor(current_batch_y).to(device), "eid":current_batch_eid, "quote_eids":current_quote_eids})

		return batches_x, batches_m, batches_y, batches_o
	

	def forward(self, batch_x, batch_m): 
		
		_, pooled_outputs, sequence_outputs = self.bert(batch_x["toks"], token_type_ids=None, attention_mask=batch_x["mask"], output_hidden_states=True, return_dict=False)

		out=sequence_outputs[-1]
		batch_size, _, bert_size=out.shape

		combined_cands=torch.matmul(batch_m["cands"],out)
		combined_quote=torch.matmul(batch_m["quote"],out)
		
		combined=torch.cat((combined_cands, combined_quote), axis=2)

		preds = self.fc(combined)
		preds=self.tanh(preds)
		preds = self.fc2(preds)

		return preds


	def evaluate(self, dev_x_batches, dev_m_batches, dev_y_batches, dev_o_batches, epoch):

		self.eval()

		cor=0.
		tot=0.
		nones=0

		gold_eids={}
		pred_eids={}

		with torch.no_grad():

			idd=0
			for x1, m1, y1, o1 in zip(dev_x_batches, dev_m_batches, dev_y_batches, dev_o_batches):
				y_pred = self.forward(x1, m1)

				orig, meta=o1
				predictions=torch.argmax(y_pred, axis=1).detach().cpu().numpy()
				for idx, pred in enumerate(predictions):

					sent=orig[idx]

					gold_eids[idd]=y1["quote_eids"][idx]

					predval=y1["eid"][idx][pred[0]]
					if predval is None:
						predval="none-%s" % (idd)

					pred_eids[idd]=predval
					val=y1["y"][idx][pred[0]]

					if pred[0] < len(meta[idx][1]):
						ent_start, ent_end, lab, ent_eid=meta[idx][1][pred[0]]

						if epoch == "test":
							print("epoch %s" % epoch, ' '.join(sent[:ent_start]), PINK, ' '.join(sent[ent_start:ent_end]), "(%s)" % int(val.detach().cpu().numpy()), ENDC, ' '.join(sent[ent_end:]))

					if val == 1:
						cor+=1
					tot+=1

					idd+=1

		precision, recall, F=b3(gold_eids, pred_eids)
		print("Nones: %s" % nones)
		print("Epoch %s, Quote F1: %.3f\tP: %.3f, R: %.3f" % (epoch, F, precision, recall))
		print("Epoch %s, accuracy: %.3f" % (epoch, cor/tot))

		return F, cor/tot








import re
import os
from collections import Counter
import sys
import argparse

from transformers import BertModel
from transformers import BertTokenizer

import torch
from torch import nn
import torch.optim as optim
import numpy as np
import random

from booknlp.common.pipelines import Token, Entity
from booknlp.english.litbank_quote import QuoteTagger
from booknlp.english.name_coref import NameCoref

from booknlp.english.bert_qa import QuotationAttribution

random.seed(1)
np.random.seed(1)
torch.manual_seed(1)
torch.backends.cudnn.deterministic = True
torch.backends.cudnn.benchmark = False

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
print("using device", device)


class BERTCorefTagger(nn.Module):

	def __init__(self, gender_cats, freeze_bert=False, base_model=None, pronominalCorefOnly=True):
		super(BERTCorefTagger, self).__init__()

		modelName=base_model
		modelName=re.sub("^coref_", "", modelName)
		modelName=re.sub("-v\d.*$", "", modelName)

		matcher=re.search(".*-(\d+)_H-(\d+)_A-.*", modelName)
		bert_dim=0
		modelSize=0

		self.num_layers=0
		if matcher is not None:
			self.num_layers=min(4, int(matcher.group(1)))
			bert_dim=int(matcher.group(2))

			modelSize=self.num_layers*bert_dim

		assert bert_dim != 0

		self.pronominalCorefOnly=pronominalCorefOnly

		self.tokenizer = BertTokenizer.from_pretrained(modelName, do_lower_case=False, do_basic_tokenize=False)
		self.bert = BertModel.from_pretrained(modelName)

		self.tokenizer.add_tokens(["[CAP]"], special_tokens=True)
		self.bert.resize_token_embeddings(len(self.tokenizer))

		self.bert.eval()

		self.vec_get_distance_bucket=np.vectorize(self.get_distance_bucket)

		if freeze_bert:
			for param in self.bert.parameters():
				param.requires_grad = False

		self.distance_embeddings = nn.Embedding(43, 20)
		self.speaker_embeddings = nn.Embedding(3, 20)
		self.nested_embeddings = nn.Embedding(2, 20)
		self.gender_embeddings = nn.Embedding(3, 20)
		self.width_embeddings = nn.Embedding(12, 20)
		self.quote_embeddings = nn.Embedding(3, 20)

		self.hidden_dim=bert_dim
		self.attention1 = nn.Linear(self.hidden_dim, self.hidden_dim)
		self.attention2 = nn.Linear(self.hidden_dim, 1)
		self.mention_mention1 = nn.Linear( (3 * self.hidden_dim + 20 + 20) * 3 + 20 + 20 + 20 + 20, 150)
		self.mention_mention2 = nn.Linear(150, 150)
		self.mention_mention3 = nn.Linear(150, 1)

		self.unary1 = nn.Linear(3 * self.hidden_dim + 20 + 20, 150)
		self.unary2 = nn.Linear(150, 150)
		self.unary3 = nn.Linear(150, 1)

		self.drop_layer_020 = nn.Dropout(p=0.2)
		self.tanh = nn.Tanh()

		self.gender_cats=gender_cats
		self.gender_expressions={}
		for val in self.gender_cats:
			cat='/'.join(val)
			self.gender_expressions[cat]={}
			for w in cat.split("/"):
				self.gender_expressions[cat][w]=1

		self.conflicting_genders={}
		for cat in self.gender_expressions:
			self.conflicting_genders[cat]={}
			for alt_cat in self.gender_expressions:
				if cat != alt_cat:
					for w in self.gender_expressions[alt_cat]:
						if w not in self.gender_expressions[cat]:
							self.conflicting_genders[cat][w]=1


	def get_mention_reps(self, input_ids=None, attention_mask=None, starts=None, ends=None, index=None, widths=None, quotes=None, matrix=None, transforms=None, doTrain=True):

		starts=starts.to(device)
		ends=ends.to(device)
		widths=widths.to(device)

		quotes=quotes.to(device)

		input_ids = input_ids.to(device)
		attention_mask = attention_mask.to(device)
		transforms = transforms.to(device)

		# matrix specifies which token positions (cols) are associated with which mention spans (row)
		matrix=matrix.to(device) # num_sents x max_ents x max_words

		# index specifies the location of the mentions in each sentence (which vary due to padding)
		index=index.to(device)

		_, pooled_outputs, sequence_outputs = self.bert(input_ids, token_type_ids=None, attention_mask=attention_mask, output_hidden_states=True, return_dict=False)

		all_layers = sequence_outputs[-1]
		embeds=torch.matmul(transforms,all_layers)

		average=torch.matmul(matrix, embeds)

		###########
		# ATTENTION OVER MENTION
		###########

		attention_weights=self.attention2(self.tanh(self.attention1(embeds))) # num_sents x max_words x 1
		attention_weights=torch.exp(attention_weights)
		
		attx=attention_weights.squeeze(-1).unsqueeze(1).expand_as(matrix)
		summer=attx*matrix

		val=matrix*summer # num_sents x max_ents x max_words
		
		val=val/torch.sum(1e-8+val,dim=2).unsqueeze(-1)

		attended=torch.matmul(val, embeds) # num_sents x max_ents x 2 * hidden_dim

		attended=attended.view(-1,self.hidden_dim)

		embeds=embeds.contiguous()
		position_output=embeds.view(-1, self.hidden_dim)
		
		# starts = token position of beginning of mention in flattened token list
		start_output=torch.index_select(position_output, 0, starts)
		# ends = token position of end of mention in flattened token list
		end_output=torch.index_select(position_output, 0, ends)

		# index = index of entity in flattened list of attended mention representations
		mentions=torch.index_select(attended, 0, index)

		average=average.view(-1,self.hidden_dim)
		averaged_mentions=torch.index_select(average, 0, index)

		width_embeds=self.width_embeddings(widths)

		quote_embeds=self.quote_embeddings(quotes)

		span_representation=torch.cat((start_output, end_output, mentions, width_embeds, quote_embeds), 1)

		if doTrain:
			return span_representation
		else:
			# detach tensor from computation graph at test time or memory will blow up
			return span_representation.detach()


	def assign_quotes_to_entity(self, entities):
		# For training, assign quotes to the nearest gold mention
		for idx, entity in enumerate(entities):
			dists=[]
			if entity.in_quote:
				for j in range(len(entities)):
					if not entities[j].in_quote and entities[j].entity_id == entity.quote_eid:
						dists.append((abs(j-idx), j))

				for dist, j in sorted(dists, reverse=False):
					entity.quote_mention=j
					break
			


	def add_property(self, entity_properties, cand_assignment, mention, ref_genders):

		if cand_assignment not in entity_properties:
			entity_properties[cand_assignment]={}

		entity_properties[cand_assignment]["ner_cat"]=mention.ner_cat
		

	def is_compatible(self, cand_mention, eid, entity_properties, mention, ref_genders, score):

		# If we've already given a referential gender to a candidate in a previous step using global 
		# information, respect that global inference when linking pronouns rather than relying on local coref to do it.
		if len(mention.text.split(" ")) == 1:
			term=mention.text.lower()

			if eid in ref_genders:
				cat=ref_genders[eid]["argmax"]
				if term in self.conflicting_genders[cat]:
					return False

		# Don't allow links between mentions with different NER categories
		if eid in entity_properties and mention.ner_cat in entity_properties[eid]:
			if mention.ner_cat != entity_properties[eid]["ner_cat"]:
				return False

		return True

	def get_non_quote_cands(self, first, idx, entities):

		# entities not in quotes can only refer back to other entities not in quotes
		dists=[]

		for i in range(first, idx):
			if entities[i].in_quote == False:
				dists.append(i)

		return np.array(dists), list(reversed(np.arange(len(dists))))


	def get_closest_entities(self, first, idx, entities, top=10):

		# entities in quotes can refer back to *any* entity (in quote or outside) or up to 10 entities outside quotes ahead
		dists=list(np.arange(first, idx))

		ent_dist=list(reversed(np.arange(len(dists))))

		k=1
		for i in range(idx+1, min(idx+1+top, len(entities))):
			
			if entities[i].in_quote == False:
				dists.append(i)
				ent_dist.append(-k)
				k+=1

		return np.array(dists), ent_dist

	def forward(self, matrix, index, existing=None, truth=None, token_positions=None, starts=None, ends=None, widths=None, input_ids=None, attention_mask=None, transforms=None, entities=None, ref_genders={}):
		
		doTrain=False
		if truth is not None:
			doTrain=True

		zeroTensor=torch.FloatTensor([0]).to(device)

		entity_properties={}

		if existing is not None:
			for idx, val in enumerate(existing):
				e=entities[idx]
				if val != -1 and val is not None:
					self.add_property(entity_properties, val, e, ref_genders)

		all_starts=None
		all_ends=None

		span_representation=None

		all_all=[]
		cur=0
		for b in range(len(matrix)):

			quotes=[]
			for entity in entities[cur:cur+len(starts[b])]:
				if entity.in_quote:
					quotes.append(1)
				else:
					quotes.append(0)
			cur+=len(starts[b])

			quotes=torch.LongTensor(quotes)

			span_reps=self.get_mention_reps(input_ids=input_ids[b], attention_mask=attention_mask[b], starts=starts[b], ends=ends[b], index=index[b], widths=widths[b], quotes=quotes, transforms=transforms[b], matrix=matrix[b], doTrain=doTrain)
			
			if b == 0:
				span_representation=span_reps
				all_starts=starts[b]
				all_ends=ends[b]

			else:

				span_representation=torch.cat((span_representation, span_reps), 0)
	
				all_starts=torch.cat((all_starts, starts[b]), 0)
				all_ends=torch.cat((all_ends, ends[b]), 0)

		all_starts=all_starts.to(device)
		all_ends=all_ends.to(device)
		
		num_mentions,=all_starts.shape

		running_loss=0

		curid=-1

		if existing is not None:
			for r in existing:
				if r > curid:
					curid=r

		curid+=1

		assignments=[None]*len(entities)

		if truth is not None:
			for idx, entity in enumerate(entities):
				assignments[idx]=entity.entity_id

		seen={}

		ch=0

		token_positions=np.array(token_positions)

		mention_index=np.arange(num_mentions)

		unary_scores=self.unary3(self.tanh(self.drop_layer_020(self.unary2(self.tanh(self.drop_layer_020(self.unary1(span_representation)))))))

		# process entities outside of quotes first

		for inQuoteVal in [False, True]:

			for i in range(num_mentions):
				
				entity=entities[i]

				if entity.in_quote != inQuoteVal:
					continue

				# if the mention has already been resolved through name coref, skip it
				if not doTrain and existing is not None and existing[i] != -1:
					assignments[i]=existing[i]
					continue

				# if the mention is a pronoun and we're only doing pronominal coref, skip it
				# (assign the mention a unique entity ID, to be clustered later)
				elif not doTrain and entity.proper != "PRON" and self.pronominalCorefOnly:

					assignment=curid
					curid+=1
					assignments[i]=assignment

					continue				

				if i == 0:
					# the first mention must start a new entity; this doesn't affect training (since the loss must be 0) so we can skip it.
					if not doTrain:
						
						if existing is None or (existing is not None and existing[i] == -1):
							assignment=curid
							curid+=1
							assignments[i]=assignment
						else:

							assignments[i]=existing[i]
					
					continue

				MAX_PREVIOUS_MENTIONS=20

				first=max(0,i-MAX_PREVIOUS_MENTIONS)

				cands_idx=None
				if inQuoteVal == False:
					# entities not in quotes can only refer back to other entities not in quotes
					cands_idx, ent_dist=self.get_non_quote_cands(first, i, entities)
				else:
					# entities in quotes can refer back to *any* entity (in quote or outside) or up to 10 entities outside quotes ahead
					cands_idx, ent_dist=self.get_closest_entities(first, i, entities)

				cands_idx=cands_idx[-MAX_PREVIOUS_MENTIONS:]
				ent_dist=ent_dist[-MAX_PREVIOUS_MENTIONS:]


				preds=None

				# if there are no candidates, then don't bother training, but create a new entity ID if predicting
				if len(cands_idx) == 0:

					if truth is None:
						
						if existing is None or (existing is not None and existing[i] == -1):
							assignment=curid
							curid+=1
							assignments[i]=assignment
						else:
							assignments[i]=existing[i]
					
					continue

				else:

					targets=span_representation[cands_idx]
					cp=span_representation[i].expand_as(targets)
					
					dists=[]
					nesteds=[]

					# force 1st person pronouns in quotes to co-refer with the quote speaker
					if entity.quote_mention is not None and assignments[entity.quote_mention ] is not None and (entity.text.lower() == "i" or entity.text.lower() == "me" or entity.text.lower() == "my" or entity.text.lower() == "myself") and entity.in_quote:
						assignments[i]=assignments[entity.quote_mention]
						continue

					same_speaker=[]
					for e in cands_idx:
						if not entities[i].in_quote:
							same_speaker.append(2)
						else:

							# the candidates for every mention in a quote have already been resolved
							# (either candidates outside of quotes, or earlier mentions within quotes)
							# check if the *speaker* of the quote the mention is embedded within is coreferent with that candidate

							if entities[i].quote_mention is None:
								same_speaker.append(2)
							else:
								attribution_assignment=assignments[entities[i].quote_mention]

								if assignments[e] == attribution_assignment:
									same_speaker.append(1)
								else:
									same_speaker.append(0)


					same_speaker_embeds=self.speaker_embeddings(torch.LongTensor(same_speaker).to(device))

					# get distance in mentions
					dists=self.vec_get_distance_bucket(ent_dist)
					dists=torch.LongTensor(dists).to(device)
					distance_embeds=self.distance_embeddings(dists)

					# is the current mention nested within a previous one?

					nest1=[]
					nest2=[]
					for cand in cands_idx:
						if entity.global_start >= entities[cand].global_start and entity.global_end < entities[cand].global_end:
							nest1.append(1)
						else:
							nest1.append(0)
						if entities[cand].global_start >= entity.global_start and entities[cand].global_end < entity.global_end:
							nest2.append(1)
						else:
							nest2.append(0)

					nesteds_embeds=self.nested_embeddings(torch.LongTensor(nest1).to(device))
					nesteds_embeds2=self.nested_embeddings(torch.LongTensor(nest2).to(device))

					elementwise=cp*targets
					concat=torch.cat((cp, targets, elementwise, distance_embeds, nesteds_embeds, nesteds_embeds2, same_speaker_embeds), 1)

					preds=self.mention_mention3(self.tanh(self.drop_layer_020(self.mention_mention2(self.tanh(self.drop_layer_020(self.mention_mention1(concat)))))))

					preds=preds + unary_scores[i] + unary_scores[cands_idx]

					preds=preds.squeeze(-1)

				if doTrain:
		
					# zero is the score for the dummy antecedent/new entity
					preds=torch.cat((preds, zeroTensor))
		
					golds_sum=0.
					preds_sum=torch.logsumexp(preds, 0)

					if len(truth[i]) == 0:
						golds_sum=0.
					else:
						golds=torch.index_select(preds, 0, torch.LongTensor(truth[i]).to(device))
						golds_sum=torch.logsumexp(golds, 0)

					# want to maximize (golds_sum-preds_sum), so minimize (preds_sum-golds_sum)
					diff=preds_sum-golds_sum

					running_loss+=diff

				else:

					if existing is not None and existing[i] != -1:
						assignments[i]=existing[i]

					elif entity.proper != "PRON" and self.pronominalCorefOnly:

						assignment=curid
						curid+=1
						assignments[i]=assignment

						continue

					else:

						assignment=None

						if i == 0 or len(cands_idx) == 0:
							assignment=curid
							curid+=1
						else:
							arg_sorts=torch.argsort(preds, descending=True)
							k=0
							while k < len(arg_sorts):
								cand_idx=arg_sorts[k]
								if preds[cand_idx] > 0:
									score=preds[cand_idx]
									
									cand_assignment=assignments[cands_idx[cand_idx]]
									cand_mention=entities[cands_idx[cand_idx]]
									if cand_assignment is None:
										print("problem!", cands_idx[cand_idx], preds[cand_idx], cand_assignment, i, inQuoteVal, assignments, torch.sort(preds, descending=True))
										sys.exit(1)
									
									if self.is_compatible(cand_mention, cand_assignment, entity_properties, entity, ref_genders, score):
										assignment=cand_assignment
										ch+=1
										self.add_property(entity_properties, cand_assignment, entity, ref_genders)
										break

								else:
									assignment=curid
									curid+=1
									break

								k+=1

						if assignment is None:
							# print("adding default new entity due to lack of compatibility", i, assignment)
							assignment=curid
							curid+=1

						assignments[i]=assignment
				

		if truth is not None:
			return running_loss
		else:
			return assignments


	def get_mention_width_bucket(self, dist):
		if abs(dist) < 10:
			return abs(dist) + 1

		return 11

	def get_distance_bucket(self, dist):

		if dist < 30:
			return dist + 10

		if dist < 40:
			return 41

		return 42

	def print_conll(self, name, sents, all_ents, assignments, out, token_maps):

		doc_id, part_id=name

		mapper=[]
		idd=0
		for ent in all_ents:
			mapper_e=[]
			for e in ent:
				mapper_e.append(idd)
				idd+=1
			mapper.append(mapper_e)

		out.write("#begin document (%s); part %s\n" % (doc_id, part_id))
		
		cur_tok=0
		tok_id=0

		for s_idx, sent in enumerate(sents):
			ents=all_ents[s_idx]
			for w_idx, word in enumerate(sent):

				if w_idx == 0 or w_idx == len(sent)-1:
					continue

				label=[]
				for idx, entity in enumerate(ents):
					start=entity.start
					end=entity.end
					if start == w_idx and end == w_idx:
						eid=assignments[mapper[s_idx][idx]]
						label.append("(%s)" % eid)
					elif start == w_idx:
						eid=assignments[mapper[s_idx][idx]]
						label.append("(%s" % eid)
					elif end == w_idx:
						eid=assignments[mapper[s_idx][idx]]
						label.append("%s)" % eid)

				out.write("%s\t%s\t%s\t%s\t_\t_\t_\t_\t_\t_\t_\t_\t%s\n" % (doc_id, part_id, tok_id, word, '|'.join(label)))
				tok_id+=1
				cur_tok+=1

				if cur_tok in token_maps[doc_id]:
					out.write("\n")
					tok_id=0

		out.write("#end document\n")


	def read_toks(self, filename):

		tok_sent_idx=0
		lastSent=None
		toks=[]
		with open(filename) as file:
			file.readline()
			for line in file:
				cols=line.rstrip().split("\t")
				parID=int(cols[0])
				sentenceID=int(cols[1])
				tokenID=int(cols[2])
				text=cols[7]
				pos=cols[10]
				lemma=cols[9]
				deprel=cols[12]
				dephead=int(cols[6])
				ner=cols[11]
				startByte=int(cols[3])

				if sentenceID != lastSent:
					tok_sent_idx=0

				tok=Token(parID, sentenceID, tok_sent_idx, tokenID, text, pos, None, lemma, deprel, dephead, ner, startByte)

				tok_sent_idx+=1
				lastSent=sentenceID
				toks.append(tok)

		return toks


	def get_matrix(self, list_of_entities, max_words, max_ents):

		matrix=np.zeros((max_ents, max_words))
		for idx, entity in enumerate(list_of_entities):
			for i in range(entity.start, entity.end+1):
				matrix[idx,i]=1
		return matrix


	def get_data(self, doc, ents, max_ents, max_words, batchsize=128):

		token_positions=[]
		ent_spans=[]
		persons=[]
		inquotes=[]

		batch_matrix=[]
		matrix=[]

		max_words_batch=[]
		max_ents_batch=[]

		max_w=1
		max_e=1

		sent_count=0
		for idx, sent in enumerate(doc):
			
			if len(sent) > max_w:
				max_w=len(sent)
			if len(ents[idx]) > max_e:
				max_e = len(ents[idx])

			sent_count+=1

			if sent_count == batchsize:
				max_words_batch.append(max_w)
				max_ents_batch.append(max_e)
				sent_count=0
				max_w=0
				max_e=0

		if sent_count > 0:
			max_words_batch.append(max_w)
			max_ents_batch.append(max_e)

		batch_count=0
		for idx, sent in enumerate(doc):
			matrix.append(self.get_matrix(ents[idx], max_words_batch[batch_count], max_ents_batch[batch_count]))

			if len(matrix) == batchsize:
				batch_matrix.append(torch.FloatTensor(np.array(matrix)))
				matrix=[]
				batch_count+=1

		if len(matrix) > 0:
			batch_matrix.append(torch.FloatTensor(np.array(matrix)))


		batch_index=[]
		batch_quotes=[]

		batch_ent_spans=[]

		index=[]
		abs_pos=0
		sent_count=0

		b=0
		for idx, sent_ents in enumerate(ents):

			for i in range(len(sent_ents)):
				index.append(sent_count*max_ents_batch[b] + i)
				#s,e,inQuote=sent[i]
				entity=sent_ents[i]
				token_positions.append(idx)
				ent_spans.append(entity.end-entity.start)
				phrase=' '.join(doc[idx][entity.start:entity.end+1])

				inquotes.append(entity.in_quote)


			abs_pos+=len(doc[idx])

			sent_count+=1

			if sent_count == batchsize:
				batch_index.append(torch.LongTensor(index))
				batch_quotes.append(torch.LongTensor(inquotes))
				batch_ent_spans.append(ent_spans)

				index=[]
				inquotes=[]
				ent_spans=[]
				sent_count=0
				b+=1

		if sent_count > 0:
			batch_index.append(torch.LongTensor(index))
			batch_quotes.append(torch.LongTensor(inquotes))
			batch_ent_spans.append(ent_spans)

		all_masks=[]
		all_transforms=[]
		all_data=[]

		batch_masks=[]
		batch_transforms=[]
		batch_data=[]

		# get ids and pad sentence
		for sent in doc:
			tok_ids=[]
			input_mask=[]
			transform=[]

			all_toks=[]
			n=0
			for idx, word in enumerate(sent):
				toks=self.tokenizer.tokenize(word)
				all_toks.append(toks)
				n+=len(toks)


			cur=0
			for idx, word in enumerate(sent):

				toks=all_toks[idx]
				ind=list(np.zeros(n))
				for j in range(cur,cur+len(toks)):
					ind[j]=1./len(toks)
				cur+=len(toks)
				transform.append(ind)

				tok_id=self.tokenizer.convert_tokens_to_ids(toks)
				assert len(tok_id) == len(toks)
				tok_ids.extend(tok_id)

				input_mask.extend(np.ones(len(toks)))

				token=word.lower()

			all_masks.append(input_mask)
			all_data.append(tok_ids)
			all_transforms.append(transform)

			if len(all_masks) == batchsize:
				batch_masks.append(all_masks)
				batch_data.append(all_data)
				batch_transforms.append(all_transforms)

				all_masks=[]
				all_data=[]
				all_transforms=[]

		if len(all_masks) > 0:
			batch_masks.append(all_masks)
			batch_data.append(all_data)
			batch_transforms.append(all_transforms)


		for b in range(len(batch_data)):

			max_len = max([len(sent) for sent in batch_data[b]])

			for j in range(len(batch_data[b])):
				
				blen=len(batch_data[b][j])

				for k in range(blen, max_len):
					batch_data[b][j].append(0)
					batch_masks[b][j].append(0)
					for z in range(len(batch_transforms[b][j])):
						batch_transforms[b][j][z].append(0)

				for k in range(len(batch_transforms[b][j]), max_words_batch[b]):
					batch_transforms[b][j].append(np.zeros(max_len))

			batch_data[b]=torch.LongTensor(batch_data[b])
			batch_transforms[b]=torch.FloatTensor(np.array(batch_transforms[b]))
			batch_masks[b]=torch.FloatTensor(batch_masks[b])
			
		tok_pos=0
		starts=[]
		ends=[]
		widths=[]

		batch_starts=[]
		batch_ends=[]
		batch_widths=[]

		sent_count=0
		b=0
		for idx, sent_ents in enumerate(ents):

			for i in range(len(sent_ents)):

				entity=sent_ents[i]

				starts.append(tok_pos+entity.start)
				ends.append(tok_pos+entity.end)
				widths.append(self.get_mention_width_bucket(entity.end-entity.start))

			sent_count+=1
			tok_pos+=max_words_batch[b]

			if sent_count == batchsize:
				batch_starts.append(torch.LongTensor(starts))
				batch_ends.append(torch.LongTensor(ends))
				batch_widths.append(torch.LongTensor(widths))

				starts=[]
				ends=[]
				widths=[]
				tok_pos=0
				sent_count=0
				b+=1

		if sent_count > 0:
			batch_starts.append(torch.LongTensor(starts))
			batch_ends.append(torch.LongTensor(ends))
			batch_widths.append(torch.LongTensor(widths))


		return batch_matrix, batch_index, token_positions, ent_spans, batch_starts, batch_ends, batch_widths, batch_data, batch_masks, batch_transforms, batch_quotes

	def get_ant_labels(self, all_doc_sents, all_doc_ents, all_quotes):

		max_words=0
		max_ents=0
		mention_id=0

		big_ents={}

		doc_antecedent_labels=[]
		quote_antecedent_labels=[]

		all_ents=[]

		big_doc_ents=[]

		for idx, sent in enumerate(all_doc_sents):
			if len(sent) > max_words:
				max_words=len(sent)

			this_sent_ents=[]
			all_sent_ents=sorted(all_doc_ents[idx], key=lambda x: (x.start, x.end))
			all_ents.extend(all_sent_ents)

			if len(all_sent_ents) > max_ents:
				max_ents=len(all_sent_ents)

			for entity in all_sent_ents:
				this_sent_ents.append(entity)

			big_doc_ents.append(this_sent_ents)


		for idx, entity in enumerate(all_ents):


			MAX_PREVIOUS_MENTIONS=20

			first=max(0,idx-MAX_PREVIOUS_MENTIONS)

			if entity.in_quote == False:
				cands_idx, _=self.get_non_quote_cands(first, idx, all_ents)

			else:
				cands_idx, _=self.get_closest_entities(first, idx, all_ents)
				
			cands_idx=cands_idx[-MAX_PREVIOUS_MENTIONS:]

			vals=[]
			for c_idx, cand_idx in enumerate(cands_idx):
				if entity.entity_id == all_ents[cand_idx].entity_id:
					vals.append(c_idx)

			doc_antecedent_labels.append(vals)

		return doc_antecedent_labels, big_doc_ents, max_words, max_ents, quote_antecedent_labels

	def read_conll(self, filename, quotes={}):

		
		sentence_breaks={}

		docid=None
		partID=None

		all_docids=[]
		all_sents=[]
		all_ents=[]
		all_antecedent_labels=[]
		all_max_words=[]
		all_max_ents=[]
		all_doc_names=[]

		all_named_ents=[]

		all_quotes=[]


		# for one doc
		all_doc_sents=[]
		all_doc_ents=[]
		all_doc_named_ents=[]
		all_doc_quotes=[]

		# for one sentence
		sent=[]
		ents={}
		sent_quotes=[]
		sent.append("[CLS]")

		# sentence ID in original CoNLL file
		sid=0

		# word ID within sentence in original CoNLL file
		wid=0

		# global token ID within document
		global_id=-1

		cur_batch_sid=0

		named_ents=[]
		cur_tokens=0
		max_allowable_tokens=425
		cur_tid=0
		open_count=0

		doc_count=0
		lastQuoteStart=None
		doc_quotes=[]

		with open(filename, encoding="utf-8") as file:


			for line in file:
				if line.startswith("#begin document"):
					doc_count+=1
					cur_batch_sid=tok_id=0

					global_id=-1
					lastQuoteStart=None
					adjusted_quotes=[]
					doc_quotes=[]

					inQuote=False, None, None, None, None, None

					all_doc_ents=[]
					all_doc_sents=[]
					all_doc_quotes=[]

					all_doc_named_ents=[]

					open_ents={}
					open_named_ents={}

					ents={}
					sent_quotes=[]
					named_ents=[]
					sent=["[CLS]"]
					cur_tokens=0
					cur_tid=0
					open_count=0

					sid=0
					wid=0

					docid=None
					matcher=re.match("#begin document \((.*)\); part (.*)$", line.rstrip())
					if matcher != None:
						docid=matcher.group(1)
						partID=matcher.group(2)

					all_docids.append(docid)

					sentence_breaks[docid]={}

				elif line.startswith("#end document"):

					all_quotes.append(doc_quotes)

					if len(sent) > 1:
						sent.append("[SEP]")
						all_doc_sents.append(sent)

						# transform ents dict to list
						ents=list(ents.values())
						ents=sorted(ents, key=lambda x: (x.start, x.end))

						named_ents=sorted(named_ents, key=lambda x: (x[0], x[1]))

						all_doc_ents.append(ents)
						all_doc_named_ents.append(named_ents)
						all_doc_quotes.append(sent_quotes)

						doc_quotes.append(adjusted_quotes)


					all_sents.append(all_doc_sents)

					doc_antecedent_labels, big_ents, max_words, max_ents, quote_antecedent_labels=self.get_ant_labels(all_doc_sents, all_doc_ents, all_doc_quotes)

					all_ents.append(big_ents)

					all_named_ents.append(all_doc_named_ents)

					all_antecedent_labels.append(doc_antecedent_labels)
					all_max_words.append(max_words+1)
					all_max_ents.append(max_ents+1)
					
					all_doc_names.append((docid,partID))

				else:

					parts=re.split("\s+", line.rstrip())

					if len(parts) < 2:
						sid+=1
						wid=0
						sentence_breaks[docid][tok_id]=1
						continue
					# new sentence
					if (cur_tokens >= max_allowable_tokens and open_count == 0):
			
						sent.append("[SEP]")
						all_doc_sents.append(sent)

						ents=list(ents.values())
						ents=sorted(ents, key=lambda x: (x.start, x.end))

						named_ents=sorted(named_ents, key=lambda x: (x[0], x[1]))
						sent_quotes=sorted(sent_quotes, key=lambda x: (x[0], x[1]))

						all_doc_ents.append(ents)
						all_doc_named_ents.append(named_ents)

						doc_quotes.append(adjusted_quotes)

						all_doc_quotes.append(sent_quotes)

						ents={}
						named_ents=[]
						sent_quotes=[]
						adjusted_quotes=[]
						sent=[]
						sent.append("[CLS]")

						cur_tokens=0

						cur_tid=0

						cur_batch_sid+=1

						if len(parts) < 2:
							continue

					# +1 to account for initial [CLS]
					tid=cur_tid+1

					token=parts[3]

					if token[0].lower() != token[0]:
						token="[CAP] " + token.lower()

					global_id+=1
					tok_id+=1
					orig_sent_id=parts[1]
					orig_token_id=parts[2]

					coref=parts[-1].split("|")
					b_toks=self.tokenizer.tokenize(token)
					cur_tokens+=len(b_toks)
					cur_tid+=1


					if docid in quotes and sid in quotes[docid]:

						# see if this word ends a quote
						for start_sid, start_wid, end_sid, end_wid, eid in quotes[docid][sid]["END"]:
							if sid == end_sid and wid == end_wid:
								inQuote=False, None, None, None, None, None
								adjusted_quotes.append((lastQuoteStart[0], lastQuoteStart[1], cur_batch_sid, tid))
			
						# see if this word starts a new quote
						for start_sid, start_wid, end_sid, end_wid, eid in quotes[docid][sid]["START"]:		
							if sid == start_sid and wid == start_wid:
								inQuote=True, eid, start_sid, start_wid, end_sid, end_wid
								lastQuoteStart=cur_batch_sid, tid

					wid+=1
					sent.append(token)

					for c in coref:
						if c.startswith("(") and c.endswith(")"):
							c=re.sub("\(", "", c)
							c=int(re.sub("\)", "", c))

							ents[(tid,tid)]=Entity(tid, tid, in_quote=inQuote[0], quote_eid=inQuote[1], quote_id=len(adjusted_quotes), entity_id=c, text=' '.join(sent[tid:tid+1]))
							ents[(tid,tid)].global_start=global_id
							ents[(tid,tid)].global_end=global_id
							
						elif c.startswith("("):
							c=int(re.sub("\(", "", c))

							if c not in open_ents:
								open_ents[c]=[]
							open_ents[c].append((tid, global_id))
							open_count+=1

						elif c.endswith(")"):
							c=int(re.sub("\)", "", c))

							assert c in open_ents

							start_tid, start_global_id=open_ents[c].pop()
							open_count-=1

							ents[(start_tid,tid)]=Entity(start_tid, tid, in_quote=inQuote[0], quote_eid=inQuote[1], quote_id=len(adjusted_quotes), entity_id=c, text=' '.join(sent[start_tid:tid+1]))
							ents[(start_tid,tid)].global_start=start_global_id
							ents[(start_tid,tid)].global_end=global_id

					ner=parts[10].split("|")

					for c in ner:

						if c.startswith("(") and c.endswith(")"):
							c=re.sub("\(", "", c)
							c=(re.sub("\)", "", c))

							if (tid, tid) in ents:
								ner_parts=c.split("_")
								ents[(tid, tid)].ner_cat=ner_parts[1]
								ents[(tid, tid)].proper=ner_parts[0]

						elif c.startswith("("):
							c=(re.sub("\(", "", c))

							if c not in open_named_ents:
								open_named_ents[c]=[]
							open_named_ents[c].append(tid)

						elif c.endswith(")"):
							c=(re.sub("\)", "", c))

							assert c in open_named_ents

							start_tid=open_named_ents[c].pop()

							if (start_tid, tid) in ents:
								ner_parts=c.split("_")
								ents[(start_tid, tid)].ner_cat=ner_parts[1]
								ents[(start_tid, tid)].proper=ner_parts[0]

		return all_sents, all_ents, all_named_ents, all_antecedent_labels, all_max_words, all_max_ents, all_doc_names, sentence_breaks, all_quotes, all_docids






"""
https://raw.githubusercontent.com/kaniblu/pytorch-bilstmcrf/master/model.py
MIT License

"""

import torch
import torch.nn as nn
import torch.nn.init as I
import torch.nn.utils.rnn as R
from torch.autograd import Variable

def log_sum_exp(vec, dim=0):
	max, idx = torch.max(vec, dim)
	max_exp = max.unsqueeze(-1).expand_as(vec)
	return max + torch.log(torch.sum(torch.exp(vec - max_exp), dim))


class CRF(nn.Module):
	def __init__(self, vocab_size, device):
		super(CRF, self).__init__()

		self.device=device
		self.vocab_size = vocab_size
		self.n_labels = n_labels = vocab_size + 2
		self.start_idx = n_labels - 2
		self.stop_idx = n_labels - 1
		self.transitions = nn.Parameter(torch.randn(n_labels, n_labels))

	def reset_parameters(self):
		I.normal(self.transitions.data, 0, 1)

	def forward(self, logits, lens):
		"""
		Arguments:
			logits: [batch_size, seq_len, n_labels] FloatTensor
			lens: [batch_size] LongTensor
		"""
		batch_size, seq_len, n_labels = logits.size()
		alpha = logits.data.new(batch_size, self.n_labels).fill_(-10000)
		alpha[:, self.start_idx] = 0
		alpha = Variable(alpha)
		c_lens = lens.clone()

		logits_t = logits.transpose(1, 0)

		for logit in logits_t:

			logit_exp = logit.unsqueeze(-1).expand(batch_size,
												   *self.transitions.size())
			alpha_exp = alpha.unsqueeze(1).expand(batch_size,
												  *self.transitions.size())
			trans_exp = self.transitions.unsqueeze(0).expand_as(alpha_exp)
			mat = trans_exp + alpha_exp + logit_exp
			alpha_nxt = log_sum_exp(mat, 2).squeeze(-1)

			mask = (c_lens > 0).float().unsqueeze(-1).expand_as(alpha)
			alpha = mask * alpha_nxt + (1 - mask) * alpha
			c_lens = c_lens - 1

		alpha = alpha + self.transitions[self.stop_idx].unsqueeze(0).expand_as(alpha)
		norm = log_sum_exp(alpha, 1).squeeze(-1)

		return norm

	def viterbi_decode(self, logits, lens):
		"""Borrowed from pytorch tutorial

		Arguments:
			logits: [batch_size, seq_len, n_labels] FloatTensor
			lens: [batch_size] LongTensor
		"""
		batch_size, seq_len, n_labels = logits.size()
		vit = logits.data.new(batch_size, self.n_labels).fill_(-10000)
		vit[:, self.start_idx] = 0
		vit = Variable(vit)
		c_lens = lens.clone()

		logits_t = logits.transpose(1, 0)
		pointers = []
		for t_idx, logit in enumerate(logits_t):
			vit_exp = vit.unsqueeze(1).expand(batch_size, n_labels, n_labels)
			trn_exp = self.transitions.unsqueeze(0).expand_as(vit_exp)
			vit_trn_sum = vit_exp + trn_exp
			vt_max, vt_argmax = vit_trn_sum.max(2)

			vt_max = vt_max.squeeze(-1)
			vit_nxt = vt_max + logit
			pointers.append(vt_argmax.squeeze(-1).unsqueeze(0))

			mask = (c_lens > 0).float().unsqueeze(-1).expand_as(vit_nxt).to(self.device)
			vit = mask * vit_nxt + (1 - mask) * vit

			mask = (c_lens == 1).float().unsqueeze(-1).expand_as(vit_nxt).to(self.device)
			vit += mask * self.transitions[ self.stop_idx ].unsqueeze(0).expand_as(vit_nxt)

			c_lens = c_lens - 1

		pointers = torch.cat(pointers)
		scores, idx = vit.max(1)
		if batch_size > 1:
			idx = idx.squeeze(-1)
		paths = [idx.unsqueeze(1)]

		for argmax in reversed(pointers):
			idx_exp = idx.unsqueeze(-1)
			idx = torch.gather(argmax, 1, idx_exp)
			idx = idx.squeeze(-1)

			paths.insert(0, idx.unsqueeze(1))

		paths = torch.cat(paths[1:], 1)
		scores = scores.squeeze(-1)

		return scores, paths

	def transition_score(self, labels, lens):
		"""
		Arguments:
			 labels: [batch_size, seq_len] LongTensor
			 lens: [batch_size] LongTensor
		"""
		batch_size, seq_len = labels.size()

		# pad labels with <start> and <stop> indices
		labels_ext = Variable(labels.data.new(batch_size, seq_len + 2))
		labels_ext[:, 0] = self.start_idx
		labels_ext[:, 1:-1] = labels
		mask = self.sequence_mask(lens + 1, max_len=seq_len + 2).long()
		pad_stop = Variable(labels.data.new(1).fill_(self.stop_idx))
		pad_stop = pad_stop.unsqueeze(-1).expand(batch_size, seq_len + 2)
		labels_ext = (1 - mask) * pad_stop + mask * labels_ext
		labels = labels_ext

		trn = self.transitions
		trn_exp = trn.unsqueeze(0).expand(batch_size, *trn.size())
		lbl_r = labels[:, 1:]
		lbl_rexp = lbl_r.unsqueeze(-1).expand(*lbl_r.size(), trn.size(0))
		trn_row = torch.gather(trn_exp, 1, lbl_rexp)
		# obtain transition score from the transition vector for each label
		# in batch and timestep (except the first ones)
		lbl_lexp = labels[:, :-1].unsqueeze(-1)
		trn_scr = torch.gather(trn_row, 2, lbl_lexp)
		trn_scr = trn_scr.squeeze(-1)

		mask = self.sequence_mask(lens, max_len=trn_scr.shape[1]).float()
		# mask = self.sequence_mask(lens + 1).float()

		trn_scr = trn_scr * mask
		score = trn_scr.sum(1).squeeze(-1)

		return score

	def _bilstm_score(self, logits, y, lens):
		y_exp = y.unsqueeze(-1)
		scores = torch.gather(logits, 2, y_exp).squeeze(-1)
		mask = self.sequence_mask(lens, max_len=scores.shape[1]).float()
		# mask = self.sequence_mask(lens).float()

		scores = scores * mask
		score = scores.sum(1).squeeze(-1)
		return score
		
	def score(self, y, lens, logits=None):

		transition_score = self.transition_score(y, lens)
		bilstm_score = self._bilstm_score(logits, y, lens)

		score = transition_score + bilstm_score

		return score

	def sequence_mask(self, lens, max_len=None):
		batch_size = lens.size(0)

		if max_len is None:
			# max_len = lens.max().data[0]
			max_len = lens.max().data


		ranges = torch.arange(0, max_len).long()
		ranges = ranges.unsqueeze(0).expand(batch_size, max_len)
		ranges = Variable(ranges)

		if lens.data.is_cuda:
			ranges = ranges.cuda()

		lens_exp = lens.unsqueeze(1).expand_as(ranges)
		mask = ranges < lens_exp

		return mask

